/*
 * xsk.c --- sk.c + skcache.c
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*

 terminology:
  entry   - object entry. store properties like control block of TCP/IP
            stack or super block of filesystem.
  frag    - fragment of object cache. object cache is devided little block
            and stored its list.
            you can access entity of object by drawing its list.
  bucket  - list of entry.  it was controled as skip list.
 
*/
/*
 NOTE:

 hash:
    this program use meny hash to assign object to structures in several level
    and identify among objects.

 on the memory:
    cache consists set of buckets and frag lists in entry.

        bucket[0]
            entry[0]    -> frag[0], frag[1], frag[2] ...
            entry[1]
            entry[2]
            ...

        bucket[1]
            entry[0]
            entry[1]
            entry[2]
            ...

 on the disk:
    cache consists dirctories according to hash.

    2level dir: (default)

        aa/bb/cc-dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc-dd - file with third and fourth hash.


    3level dir: (optional)

        aa/bb/cc/dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc - third level dir. with third hash
        dd - file with fourth hash.

 */

/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 * $Id$
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * The algorithm on the original text was described by Pascal like
 * langauage.  Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>

#include <dirent.h>

#include <malloc.h>
#include <sys/mman.h>

#include <assert.h>
#include <errno.h>


#include "range.h"
#include "msg.h"
#include "misc.h"
#include "mem.h"

#include "xsk.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "mlist.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DISK_RETRY
#define DISK_RETRY (13)
#endif

#ifndef FRAG_NTRY
#define FRAG_NTRY   (32)
#endif


/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

int radix_h1 =   241;
int radix_h2 =   239;
int radix_h3 =   233;
int radix_h4 =   251;

skcache_entry_t *NIL;
int skcache_bucket_num;
bucket_t *skcache_bucket_base;

extern pthread_mutex_t stat_lock;

volatile int swapload_count     = 0;
volatile int swapload_success   = 0;
volatile int swapin_count       = 0;
volatile int swapin_success     = 0;
volatile int swapout_count      = 0;
volatile int swapout_success    = 0;

volatile int checkdir_dir_count         = 0;
volatile int checkdir_file_read_count   = 0;
volatile int checkdir_file_handle_count = 0;
volatile int checkdir_file_left_count   = 0;
volatile int checkdir_file_delete_count = 0;
volatile int checkdir_file_erase_count  = 0;

volatile int check_dir_ignore_hash_count = 0;

sim_stat_rec    swapload_time;
static long     swapload_time_count[K(1)];
sim_stat_rec    swapin_time;
static long     swapin_time_count[K(1)];
sim_stat_rec    swapout_time;
static long     swapout_time_count[K(1)];
sim_stat_rec    compaction_time;
static long     compaction_time_count[K(2)];
sim_stat_rec    checkdir_time;
static long     checkdir_time_count[K(2)];
sim_stat_rec    checkdir_time_short;
static long     checkdir_time_short_stat_count[K(2)];
sim_stat_rec    checkmem_time;
static long     checkmem_time_count[K(2)];

sim_stat_rec    swapout_fail_reason;
static long     swapout_fail_reason_count[K(1)];
sim_stat_rec    swapload_fail_reason;
static long     swapload_fail_reason_count[K(1)];



int require_radix_h2;
int require_radix_h3;
off_t require_dir_size;
int require_swap_token_val;
int require_swap_token_interval;
int require_disk_sweep_interval;
int require_keeptime_token_val = -1;



extern pthread_mutex_t config_lock;

int skcache_keeptime_token_init = 10;
int skcache_keeptime_token      = 10;
int skcache_mem_keep_min    = 12;
int skcache_mem_keep_max    = 3600*4;   /* 4 hour, 1/6 day */
int skcache_mem_keep_cur    = 3600;
int skcache_entry_keep_min  = 120;
int skcache_entry_keep_cur  = 86400;    /* 1 day */
int skcache_entry_keep_max  = 86400*7;  /* 1 week */

double skcache_frag_hwater_ratio = 80.0;
double skcache_frag_lwater_ratio = 60.0;

int skcache_signature_refresh = DSIG_REFRESH;

volatile long long swap_count = 0LL;
volatile long long swap_size = 0LL;

/* limitter */
volatile int swapout_token_init = 1000;
volatile int swapout_token  = 1000;
volatile int swapout_token_interval;
volatile int disk_sweep_interval;

time_t min_date;

pthread_mutex_t dir_mutex = PTHREAD_MUTEX_INITIALIZER;
int dir_num=0;
int *dir_lock;
/*
int *dir_sig;
int64_t *dir_sig;
*/
u_int32_t *dir_sig;
off_t *dir_sz;
int *dir_count;
int *dir_time;
int dir_conflict = 0;

pthread_mutex_t frag_lock = PTHREAD_MUTEX_INITIALIZER;
skcache_frag_t *z_frag_freelist = NULL;
int frag_use = -1, frag_free = -1;
int frag_hhwater, frag_hwater, frag_lwater;


sim_hist_rec fratio_hist;
static long fratio_hist_area[128];


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, msleep is made.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;

    w.tv_sec = 0;
    w.tv_nsec = x * 1000000;   /* convert msec to nsec */

    nanosleep(&w, NULL);

    return 0;
}
#endif



void
sk_rawdump(FILE* fp, skiplist_t *list, char* (fn)(void*))
{
    skiplist_cell_t *cur;
    int c, i;

#if 0
    for(i=0;i<list->level;i++) {
        fprintf(stdout, " h %4d: %p\n", i, list->header->skip[i]);
    }
#endif

    fprintf(fp, "--- list addr %p, num %d, func %p\n",
        list, list->num, fn);

    c = 0;
    cur = list->header->skip[0];
    if(fn)
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p %s // %p\n",
                c, cur, cur->entity, fn(cur->entity), cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
    else
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p // %p\n",
                c, cur, cur->entity, cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
}


void
sk_dump(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

    fprintf(fp, "%-24s %-10s ", "header", "header");
    for(i=0;i<=list->level;i++) {
        if(list->header->skip[i]==NULL) {
            fprintf(fp, "%-10s ", "NIL");
        }
        else {
            fprintf(fp, "0x%08x ", list->header->skip[i]);
        }
    }
    fprintf(fp, "\n");

#if 1
    cur = list->header->skip[0];
    while(cur) {
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
        }
        else {
            strcpy(label, "*NONE*");
        }


        fprintf(fp, "%-24s 0x%08x ", label, cur);

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-10s ", "NIL");
            }
            else {
                fprintf(fp, "0x%08x ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

        if(cur==list->term) {
            break;
        }
        if(cur==cur->skip[0]) {
            fprintf(fp, "*** looped\n");
            break;
        }

        cur = cur->skip[0];
    }

    fprintf(fp, "%-24s %-10s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
#endif

}

void
bk_dump(bucket_t *bk, skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    char *pcur;
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int j, k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d, term %p\n",
        list, list->num, list->level, list->maxlevel, list->term);

#ifndef DEBUG
    return;
#endif

#ifdef DEBUG
    fprintf(fp, "#  first %p last %pd\n#\n",
        bk->cell_first, bk->cell_last);

#if 0
    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %s %p\n",
            "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s ",
            "", "", "");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %s %p\n",
            "", "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s %-14s ",
            "", "", "", "");
    }
    if(list->level<0) {
        fprintf(fp,"*** no level ***\n");
    }
    else {

        cur = list->header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
    fprintf(fp, "\n");

    }
#endif

#if 1
    fprintf(fp, "# cup  header %p: ", bk->cup.header);
        cur = bk->cup.header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
    fprintf(fp, "# cup  term   %p: ", bk->cup.term);
        cur = bk->cup.term;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

#endif


    fprintf(fp, "#\n");

    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %8s",
            "num", "addr", "entity", "skips..");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %8s",
            "num", "addr", "entity", "label", "skips..");
    }
    fprintf(fp, "\n");

    j = -1;
    k = 0;
    pcur = (char*)bk->cell_first;
    while(pcur<=(char*)bk->cell_last) {
        cur=(skiplist_cell_t*)pcur;

        j++;
        if(cur->entity==NULL) {
            goto next;
        }
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
            fprintf(fp, "= %6d %6d %p %p %-14s ",
                j, k, cur, cur->entity, label);
        }
        else {
            fprintf(fp, "= %6d %6d %p %p ",
                j, k, cur, cur->entity);
        }

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "x");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
        k++;
#if 0
        if(k>list->num+3) {
            break;
        }
#endif

next:
        pcur += bk->elemsize;
    }

    fprintf(fp, "\n\n");
    fflush(fp);

#endif
}




static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 4097;
    double v;

    lv = 0;
#if 0
    while( v=((double)(rand()%w))/w, v < p && lv < max) {
        lv++;
    }
#endif
    while(drand48() < p && lv < max) {
        lv++;
    }

    return lv;
}




/*
 * cmp - compare
 */
skiplist_cell_t *
sk_find_listpos(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    skiplist_cell_t *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i];
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x;
    }

out:
    return ret;
}

void*
sk_find(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    void *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i]->entity;
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x->entity;
    }

out:
    return ret;
}


/*
 * if alreay assigned, not new allocation.
 * cmp - compare
 */
void*
sk_assign(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*),
        int* st, int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    skiplist_cell_t *ret;
    int c;
    int diff;

#if 1
    if(newface==NULL) {
        Error("don't assign NULL to skiplist\n");

        ret = NULL;
        *st = -1;
        goto out;
    }
#endif

    if(*flist==NULL) {
        fprintf(stderr, "sk_assign: no free cell in entity list (opt %d)\n", opt);
        ret = NULL;
        *st = -1;
        goto out;
    }

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term
                        || x->skip[i]->entity==NULL) {
                diff = INT_MAX;
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);

            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
        *st = 1;
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            ret = NULL;
            *st = -1;
            goto out;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        ret = x->entity;
        *st = 0;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

out:
    return ret;
}

int
sk_insert(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (ovw)(void**,void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[0]==NULL || x->skip[0]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0 && ovw) {
        ovw(&x->entity, newface);
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            return -1;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

    return 0;
}

/*
 * arguments:
 *      cmp - compare
 *      rem - remove
 * return:
 *      NULL    fail    
 *      !NULL   success
 */
void*
sk_delete(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (rem)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
    int i;
    void *ret;
    int diff;
    int chk;
    int z;
    char* skcache_label4_ref(void *pp);

    ret = NULL;

    x = list->header;
    z = 0;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term) {
                break;
            }
            if(x->skip[i]->entity==NULL) {
                Error("sk_delete: found empty entity");
#if 1
    sk_dump(list, stderr, skcache_label4_ref);
#endif
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        update[i] = x;
    }
    x = x->skip[0];

    if(x==NULL) {
Error("sk_delete: meat NULL element");
        return ret;
    }
    if(x==list->term) {
Error("sk_delete: reach to end");
        return ret;
    }
    if(x->entity==NULL) {
Error("sk_delete: meat NULL entity");
        return ret;
    }

    if(x->entity == (skcache_entry_t*)newface
      || cmp(x->entity, newface)==0) {
        if(rem) {
            chk = rem(x->entity);
            if(chk) {
Error("sk_delete: rem() return !0 (%d)", chk);
                return ret;
            }
        }

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x) {
                update[i]->skip[i] = x->skip[i];
            }
        }

        while(list->level>0 &&
                list->header->skip[list->level]==list->term) {
            list->level--;
        }

        if(x->entity==NULL) {
Error("sk_delete: entity is NULL");
        }

        ret = x->entity;
        x->entity = NULL;
        for(i=1;i<=list->maxlevel;i++) {
            /*
            x->skip[i] = NULL;
            */
            x->skip[i] = list->term;
        }
        x->skip[0] = *flist;
        *flist = x;

        list->num--;
    }
#if 1
    else {
        Trace("sk_delete: not found entry %p\n", newface);
    }
#endif

#if 0
    if(ret==NULL) {
Error("sk_delete: return is NULL");
    }
#endif

    return ret;
}

/*
 * ack - decide delete or not
 *          that function have to return not zero for delete
 */
int
sk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */

        if(ack==NULL)
            goto delete;

        chk = ack(x->entity, opt);

        if(chk) {
delete:
            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
#if 0
                y->skip[i] = NULL;
#endif
                y->skip[i] = list->term;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }
no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}

skiplist_t*
sk_new(int nlevel, void* user_term)
{
    skiplist_t *this;
    skiplist_cell_t *n;
    int i, j;

    this = (skiplist_t*) malloc(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }
    MLOCK((void*)this, sizeof(skiplist_t)*1);
    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;
    n = this->header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        MLOCK((void*)n->skip, 
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }
    n = this->term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        MLOCK((void*)n->skip,
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    return this;
}


int
bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    return 0;
}

int
bk_new(bucket_t *bk, int nlevel, int nelement, void* user_term)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;
    char *pool;


    if(bk==NULL) {
        return 0;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    bk->elemnum = nelement;
    bk->elemsize = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(nlevel+1);
    bk->totalsize = bk->elemsize*(nelement+4);
    pool = (char*)malloc(bk->totalsize);
    if(pool==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        return -1;
    }
    MLOCK((void*)pool, bk->totalsize);

#if 0
    fprintf(stderr, "nlevel %d, nelement %d\n",
            nlevel, nelement);
    fprintf(stderr, "element size %d, num %d -> total %d\n",
            bk->elemsize, nelement, bk->totalsize);
    fprintf(stderr, "pool  [%p..%p)\n", pool, pool+bk->totalsize);
#endif
#if 0
    for(i=0;i<nelement+2;i++) {
        fprintf(stdout, "%4d: <%p>\n", i, pool+bk->elemsize*i);
    }
    fflush(stdout);
#endif
    if(pthread_mutex_init(&bk->mutex, NULL)<0) {
        Error("fail pthread_mutex_init() (%d)", errno);
    }

    bk->cup.num         = 0;
    bk->cup.maxlevel    = nlevel;
    bk->cup.level       = 0;
    bk->cup.header      = (skiplist_cell_t*)pool;
    bk->cup.term        = (skiplist_cell_t*)(pool + 1*bk->elemsize);
    n = bk->cup.header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }
    n = bk->cup.term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }


#if 0
    fprintf(stderr, "cup  header %p\n", bk->cup.header);
    fprintf(stderr, "cup  term   %p\n", bk->cup.term);
#endif

    bk->elem_freelist = (skiplist_cell_t*)(pool+4*bk->elemsize);
#if 0
    fprintf(stderr, "elem_freelist    %p\n", bk->elem_freelist);
#endif


#if 0
#ifndef NDEBUG
    bk->cell_first = (skiplist_cell_t*)(pool+0*bk->elemsize);
    bk->cell_last  = (skiplist_cell_t*)(pool+nelement*bk->elemsize);

    fprintf(stderr, "  first..last %p..%p\n", bk->cell_first, bk->cell_last);
#endif
#endif

    n = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL) {
            break;
        }
#if 0
        fprintf(stdout, "* %d, %p\n", i, n);
        fflush(stdout);
#endif
        n->entity = NULL;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=1;j<=nlevel;j++) {
/*
            n->skip[j] = NULL;
*/
            n->skip[j] = bk->cup.term;
        }
        if(i<nelement-1) {
            n->skip[0] = (skiplist_cell_t*)((char*)n+bk->elemsize);
            n = (skiplist_cell_t*)((char*)n+bk->elemsize);
        }
        else {
            n->skip[0]  = NULL;
        }
    }

#if 0
    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    fprintf(stdout, "free: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL)
            break;

#if 0
        fprintf(stdout, "%4d: <%p> ", i, n);
        for(j=0;j<=nlevel;j++) {
            fprintf(stdout, "%d:<%p> %p ",
                j, &(n->skip[j]), n->skip[j]);
        }
        fprintf(stdout, "\n");
#endif
#if 1
        fprintf(stdout, "%4d: <%p> %p\n", i, n, n->skip[0]);
#endif

        n = n->skip[0];
    }

#endif

    return bk->totalsize;
}


int
check_emptyentity(bucket_t *bk)
{
    int ret;
    skiplist_t *list;
    register skiplist_cell_t *x;

    ret = 1;
    list = &bk->cup;
    x = list->header;
    x = x->skip[0];
    while(x!=list->term) {
        if(x->entity == NULL) {
            Error("check_emptyentity: bk# %d, elem %p have empty entity",
                bk->bid, x);
            ret = 0;
            goto out;
        }
        x = x->skip[0];
    }

out:
    return ret;
}


#include "xsk_hash.c"
extern int calc_hash(unsigned char *b, int *h1, int *h2, int *h3, int *h4);


void
dsig_mkhumanstr(u_int32_t *s, char *line)
{
    int i;
    u_int32_t m, v;
    char *p;

    p = line;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        m = 1U << i%(SIZE_OF_UINT*8);
        v = s[i/(SIZE_OF_UINT*8)] & m;
        *p++ = (v ? '1' : '0');
    }
    *p = '\0';
}

void
dsig_dump(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */

    dsig_mkhumanstr(s, line);
    Trace("%s\n", line);
}

void
dsig_show(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */
    int i, c;

    dsig_mkhumanstr(s, line);
    c = 0;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        if(DSIG_TMP_ISSET(s, i)) {
            c++;
        }
    }
    Trace("addr %p count %d, bits %s\n", s, c, line);
}


void
dsig_set(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    s[i/(SIZE_OF_UINT*8)] |= m;
}

int
dsig_isset(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    return (s[i/(SIZE_OF_UINT*8)] & m) ? 1 : 0;
}

int
dsig_allset(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)-1L; /* all bits were set */
        s[i] = u;
    }
}

int
dsig_allclear(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)0UL;
        s[i] = u;
    }
}




int
skcache_entry_show(skcache_entry_t* e, FILE *ofp)
{
    fprintf(ofp,"\
-/-/\n\
entry %p, oid# %d, hash %d:%d:%d:%d\n\
name %s\n\
count %d, ref %d, cracked %d, owner %p\n\
issue %d, hot %d, lmd %d, expire %d\n\
totallength %lld\n\
mem record %p, disk record %d\n\
\n",
        e, e->oid, e->h1, e->h2, e->h3, e->h4,
        e->name,
        e->count, e->ref, e->cracked,
        e->owner, 
        e->issue_time, e->hot, e->lmd, e->expire,
        (long long)e->totallength,
        e->m_recorded, e->d_recorded);
    
    return 0;
}


int
skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            if(x->h3==y->h3) {
                return strcmp((const char*)x->name, (const char*)y->name);
            }
            else {
                return x->h3 - y->h3;
            }
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

int
X_skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            return x->h3 - y->h3;
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

char*
skcache_label_hash(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_PATH];

    sprintf(tmp, "%04d:%04d:%04d", p->h1, p->h2, p->h3);

    return tmp;
}

char*
skcache_label_short(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d %s", p->h1, p->h2, p->h3, p->name);

    return tmp;
}

char*
skcache_label_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d-%03d", p->h1, p->h2, p->h3, p->ref);

    return tmp;
}

char*
skcache_label4_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d:%04d-%03d",
        p->h1, p->h2, p->h3, p->h4, p->ref);

    return tmp;
}


int
skcache_dump_entry_frag(char *msg, skcache_entry_t *en)
{
    skcache_frag_t *p;
    off_t frag_sum;
    int u;

    Trace("%s\n", msg);
    Trace("  frags of entry %p\n", en);
    if(en==NULL)
        return -1;

    p = en->m_recorded;
    frag_sum = 0L;
    u = 0;
    while(p) {
        frag_sum += p->len;
        Trace("  %4d: p %p next 0x%08x | len %4d/%d %7ld\n",
            u, p, p->next, p->len, p->size, (long)frag_sum);
        if(p==p->next) {
            Trace("  looped !!!\n");
            break;
        }
        p = p->next;
        u++;
    }
    Trace("  num %d, totalength %lld\n",
        u, (long long)en->totallength);

    return u;
}


off_t
skcache_show_with_HTML(int ofd)
{
    int i;
    off_t wsum, w;
    skiplist_t      *list;
    skiplist_cell_t *cur;
    skcache_entry_t *en;

    wsum = 0;

    w = fdprintf(ofd, "<html>\n");
    wsum += w;



    for(i=0;i<skcache_bucket_num;i++) {
        w = fdprintf(ofd, "<p>bucket# %d\n", i);
        wsum += w;

        w = fdprintf(ofd, "<listing>\n");
        wsum += w;

        list = &skcache_bucket_base[i].cup;
        cur = list->header->skip[0];

        while(cur && cur!= list->term) {
            en = (skcache_entry_t*)cur->entity;
            w = fdprintf(ofd, " %04d:%04d:%04d %d %d '%s'\n",
                    en->h1, en->h2, en->h3, en->ref, en->count, en->name);
            if(w>0) {
                wsum += w;
            }

            cur = cur->skip[0];
        }

        w = fdprintf(ofd, "</listing>\n");
        wsum += w;

    }


    w = fdprintf(ofd, "<html>\n");
    wsum += w;

    Trace("write %ld bytes\n", wsum);
}


int
skcache_have(int fh, int sh, int th, int ff, char *name,
    int flag_check_rcode, time_t ref_time)
{
    int ret;
    skcache_entry_t tmp;
    skcache_entry_t *stored_pos;
    bucket_t *bk;


    ret = 0;

    bk = &skcache_bucket_base[fh];

    tmp.h1 = fh;
    tmp.h2 = sh;
    tmp.h3 = th;
    tmp.h4 = ff;
    strcpy(tmp.name, name);

    LOCK_BK(bk);

    stored_pos = (skcache_entry_t*) sk_find(&bk->cup, &tmp, skcache_entry_cmp);

    if(stored_pos) {
#if 1
        if(strcmp((const char*)stored_pos->name, (const char*)name)!=0) {
            Error("skcache_have: hash collision %04d:%04d:%04d '%s' v.s. '%s'",
                fh, sh, th, stored_pos->name, name);
            goto out;
        }
#endif

        if(flag_check_rcode) {
        }

        if(ref_time>0) {
        }

        ret = 1;
    }
    else {
        ret = 0;
    }

out:
    UNLOCK_BK(bk);

    return ret;
}


static
long long
alloc_entry_freelist(skcache_entry_t** head, int num)
{
    char *base;
    int esize;
    int i;
    int ret;
    skcache_entry_t *pos;
    skcache_entry_t *last;

    esize = sizeof(skcache_entry_t);
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }
    MLOCK((void*)base, esize*num);

    for(i=0; i<num; i++) {
        pos = (skcache_entry_t*)(base + esize*i);

        pos->magic[0] = MAGIC_0;
        pos->magic[1] = MAGIC_1;

        pos->oid = -1;
        pos->ref = -1;
        pos->h1 = -1;
        pos->h2 = -1;
        pos->h3 = -1;
        pos->name[0] = '\0';
        pos->m_recorded = NULL;
        pos->d_recorded = -1;

        pos->next = (skcache_entry_t*)((char*)pos + esize);
    }

    *head = (skcache_entry_t*)base;
    pos->next = NULL;

#if 0

    Trace("entry// head %p, num %d\n", head, num);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
                i, pos, pos->name, pos->m_recorded, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}

static
long long
alloc_frag_freelist(skcache_frag_t** head, int num, int sz)
{
    char *base;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;

    esize = sizeof(skcache_frag_t)+sz;
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }
    MLOCK((void*)base, esize*num);

    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

#if 0
    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    i = 0;
    pos = *head;
    while(pos) {
        if(i>num-3) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);
        }

        pos = pos->next;
        i++;
    }

#endif

#if 0
    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}


inline
int
skcache_entry_check(skcache_entry_t* e, char*f, int l)
{
    if(e==NULL) {
        Error("echk: %s %d : empty entry", f, l);
        return -1;
    }
#ifdef CHECK_MAGIC
    if(e->magic[0]!=MAGIC_0 || e->magic[1]!=MAGIC_1) {
        Error("echk: %s %d : entry %d, oid# %d, ignore magic",
            f l, e, e->oid);
        return 2;
    }
#endif
#if 1
    if(e->ref<0) {
        Error("echk: %s %d : entry %d, oid# %d, neg ref",
            f, l, e, e->oid);
        return 3;
    }
#endif
    return 0;
}

int skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);
int RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);

int
multi_frag_lock()
{
    int chk;
    int c;

    c = FRAG_NTRY*4;
try_lock:
    c--;
    chk = MUTEX_TRYLOCK(&frag_lock);
    if(chk==EBUSY) {
        if(c>0) {
#if 0
            Trace("skcache_entry_purge: free list is busy, try %d, u %d, f %d\n",
                c, frag_use, frag_free);
#endif
            msleep(3);
            goto try_lock;
        }
        Error("skcache_entry_purge: free list is busy, give up, u %d, f %d",
            frag_use, frag_free);
        return -1;
    }

    return 0;
}

int
skcache_entry_purge(bucket_t *bk, skcache_entry_t *e)
{
    skcache_frag_t *cur, *pre;
    int chk;

    chk = multi_frag_lock();
    if(chk) {
        return -1;
    }

    if(e->m_recorded) {
        /* remove */
        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            RAW_skcache_push_idle_frag(bk, cur);
            bk->frag_use--;
            if(pre==cur) {
                Error("skcache_entry_purge: bk %p, entry %p, frag %p is looped",
                    bk, e, cur);
                break;
            }

            cur = pre;
        }

        e->m_recorded = NULL;
    }

    MUTEX_UNLOCK(&frag_lock);

    return 0;
}



int
skcache_entry_delete(bucket_t *bk, skcache_entry_t *e)
{
    void *chkv;
    int chk;
    int ret;

    if(ret = skcache_entry_check(e, __FILE__, __LINE__)) {
        goto out;
    }
    ret = 0;

#if 0
    if(e->ref!=0)
    Trace("skcache_entry_delete: bk# %d %p, entry %p, oid# %d, ref %d\n",
        bk->bid, bk, e, e->oid, e->ref);
#endif

 assert(check_emptyentity(bk));
    chk = skcache_entry_purge(bk, e);
 assert(check_emptyentity(bk));
    if(chk) {
        Error("skcache_entry_delete: fail skcache_entry_purge() bk# %d, entry %x",
            bk->bid, e);
        ret = 1;
    }
    else {
 assert(check_emptyentity(bk));
        chkv = sk_delete(&bk->cup, e,
            &bk->elem_freelist, skcache_entry_cmp, NULL, 0);
 assert(check_emptyentity(bk));
        if(chkv==NULL) {
#if 0
            Trace("skcache_entry_delete: bk# %d, entry %p is not found\n",
                bk->bid, e);
#endif 

            ret = 2;

            chkv = sk_find(&bk->cup, e, skcache_entry_cmp);
            Trace("  sk_find() returns %x\n", chkv);

        }
        else {
            chk = skcache_push_idle_entry(bk, e);
#if 1
            if(chk) {
#if 0
                Trace("skcache_entry_delete: chk %d\n", chk);
#endif
                ret = 3;
            }
#endif
        }
    }

out:
    return ret;
}


#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif


char* cache_dir_prefix = "/var/tmp/cache";
int radix_disk = -1;


inline
char*
assign_dir_prefix(char *name, int d1)
{
    if(radix_disk<=0) {
        sprintf(name, "%s/%d", cache_dir_prefix, d1);
    }
    else {
        sprintf(name, "%s%d/%d",
            cache_dir_prefix, d1%radix_disk, d1);
    }

    return name;
}


inline
char*
assign_dirname_bynumber(char *name, int d1, int d2, int d3)
{
    char tmp[L_PATH];

    assign_dir_prefix(name, d1);

#ifdef USE_3LEVEL_DIR
    sprintf(tmp, "/%d/%d", d2, d3);
#else
    sprintf(tmp, "/%d", d2);
#endif
    strcat(name, tmp);

    return name;
}

inline
char*
assign_dirname(char* name, skcache_entry_t *e, int level)
{
    assign_dirname_bynumber(name, e->h1, e->h2, e->h3);

    return name;
}

inline
char*
assign_filename(char *name, skcache_entry_t *e)
{
    char tmp[L_PATH];

    assign_dirname(name, e, -1);
    
#ifdef USE_3LEVEL_DIR
    sprintf(tmp, "/%d", e->h4);
#else
    sprintf(tmp, "/%d-%d", e->h3, e->h4);
#endif
    strcat(name, tmp);

    return name;
}



int
build_dir_bynumber(int d1, int d2, int d3)
{
    int i;
    int ret=0;
    int chk;
    char name[L_PATH];
    char tmp[L_PATH];

l1:
    assign_dir_prefix(name, d1);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 1 (%d)", name, errno);
        ret = -1;
        goto out;
    }

l2:
    sprintf(tmp, "/%d", d2);
    strcat(name, tmp);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 2 (%d)\n", name, errno);
        ret = -1;
        goto out;
    }

#ifdef USE_3LEVEL_DIR
l3:
    sprintf(tmp, "/%d", d3);
    strcat(name, tmp);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 3 (%d)\n", name, errno);
        ret = -1;
        goto out;
    }
#endif

out:
    return ret;
}

int
build_dir(skcache_entry_t *e, int lev)
{
    int ret=0;

    ret = build_dir_bynumber(e->h1, e->h2, e->h3);

out:
    return ret;
}




/*
 * return
 *      0   success
 *      >0  fail
 *
 */

int
skcache_entry_swapload(bucket_t *bk, skcache_entry_t *e)
{
    int sfd;
    char name[L_PATH];
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre;
    int nr;
    off_t nr_sum;
    int chk;
    char *namechk;
    int i;
    int ret;
    int nf;
    struct timeval begin, end;
    int c;
    off_t fsize;
    off_t tmpfsize;
    int erase;
    int sigpos;
    int lf;


    gettimeofday(&begin, NULL);

    sigpos = DSIG_POS(e->h1, e->h2, e->h3);

    MUTEX_LOCK(&stat_lock);
    swapload_count++;
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&stat_lock);

    ret = 0;

#if 0
    Trace("skc_swapload: p %p, oid# %d, bk# %d, %p\n", e, e->oid, bk->bid, bk);
#endif
#if 0
    Trace("skc_swapload: p %p, id# %d, bk %p, frag = %4d/%d, len %ld (%% %d)\n",
        e, e->oid,
        bk, bk->frag_use, bk->frag_all,
        (long)e->totallength, e->totallength%1460);
#endif
#if 0
    if(e->d_recorded<0) {
        Error("swapload: not recorded in disk bk# %d, e %p", bk->bid, e);
        ret = 1;
        goto out;
    }
#endif
    if(e->m_recorded!=NULL) {
        Error("swapload: not recorded in memory bk# %d, e %p", bk->bid, e);
        ret = 2;
        goto out;
    }
#if 0
    if(bk->frag_use>bk->frag_all) {
        Error("swapload: no frag in freelist");
        ret = 3;
        goto out;
    }
#endif

#if 0
    if(bk->dirty2 == e->h2 && bk->dirty3 == e->h3) {
        Trace("swapload: conflict directory [%03d:%03d:%03d] to sweeper\n",
            e->h1, e->h2, e->h3);
        ret = 73;
        goto out;
    }
#endif

    if(lf>0) {
/*
        Trace("swapload: conflict directory bk# %d, dir# %d,.\n", bk, sigpos);
*/
        MUTEX_LOCK(&dir_mutex);
        dir_conflict++;
        MUTEX_UNLOCK(&dir_mutex);

        ret = 73;
        goto out;
    }


    namechk = assign_filename(name, e);

#if 0
    Trace("  name '%s'\n", name);
#endif
    sfd = open(name, O_RDONLY, 0666);
    if(sfd<0) {
#if 0
        Trace("swapload: cannot open swapload file '%s' (%d)\n", name, errno);
#endif
#if 0
        Error("cannot open swapload file '%s' (%d)", name, errno);
#endif
        ret = 4;
        goto out;
    }

try_lseek0:
    fsize = (off_t) lseek(sfd, 0, SEEK_END);
    if(fsize<0) {
        if(errno==EINTR) {
            goto try_lseek0;
        }

        ret = 192;
        goto fail_read_cntl;
    }
    if(fsize<sizeof(skcache_entry_t)) {
#if 0
        Trace("swapload: file size too short %ld bytes\n", (long)fsize);
#endif
        ret = 122;
        goto fail_read_cntl;
    }
    if(fsize>(off_t)0) {
try_lseek1:
        tmpfsize = lseek(sfd, 0, SEEK_SET);
        if(tmpfsize<0 && errno==EINTR) {
            goto try_lseek1;
        }
    }

    c = 0;
read_cntl:
    c++;
    nr = read(sfd, &tmp_entry, sizeof(skcache_entry_t));
    if(nr<0) {
        if(errno==EINTR && c < DISK_RETRY) {
            goto read_cntl;
        }
        Error("swapload: cntl, fail read() return %d %c (%d)", nr, c, errno);
        ret = 123;
        goto fail_read_cntl;
    }

    if(nr<sizeof(skcache_entry_t)) {
#if 0
        Trace("swapload: read size too short %ld bytes\n", (long)nr);
#endif
        ret = 126;
        goto fail_read_cntl;
    }

    /*
     * this routing MUST check magic.
     * because disk maybe stores broken or older version cache file. 
     */
    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
#ifdef CHECK_MAGIC
        Error("swapload: ignore magic %02x %02x / %02x %02x '%s'",
            tmp_entry.magic[0], tmp_entry.magic[1],
            MAGIC_0, MAGIC_1, name);
#endif
        ret = 125;
        goto fail_read_cntl;
    }

    tmp_entry.m_recorded = NULL;
    tmp_entry.d_recorded = tmp_entry.oid;

#if 0
    Trace("swapload: read oid# %d, hash %d,%d,%d,%d %s\n",
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
#endif

    if(tmp_entry.oid<0) {
#if 0
        Trace("swapload: reject negative oid to %p\n", e);
#endif
        ret = 127;
        goto fail_read_cntl;
    }

    if(strcmp(tmp_entry.name,e->name)!=0) {
#if 0
        Trace("swapload: swapload file '%s'\n", name);
        Trace("swapload: read oid# %d, hash %d,%d,%d,%d %s\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
        Trace("but target is '%s'\n", e->name);
#endif

#if 0
        Trace("swapload: swapload file '%s'\n", name);
        Trace("    oid# %d, hash %d,%d,%d,%d\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4);
        Trace("    saved %s\n    wont  %s\n", tmp_entry.name, e->name);
        {
            int yh1, yh2, yh3, yh4;
            calc_hash(tmp_entry.name,   &yh1, &yh2, &yh3, &yh4);
            calc_hash(e->name,          &yh1, &yh2, &yh3, &yh4);
        }
#endif

        ret = 133;
        erase = 1;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<0) {
#if 1
        Error("swapload: ignore issue_time");
#endif
        ret = 134;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<min_date) {
#if 1
        Error("swapload: very old object");
#endif
        ret = 129;
        goto fail_read_cntl;
    }

    /* XXX */
    *e = tmp_entry;

skip_read_cntl:
    (void)1;

    pre = NULL;

    nr_sum = 0L;
    nf = 0;     /* the number of frag */
    do {
        cur = skcache_pull_new_frag(bk);
        if(cur==NULL) {
            ret = 7;
            goto fail_read;
        }
        bk->frag_use++;

        c=0;
try_read:
        c++;
        nr = read(sfd, cur->ent, cur->size);
#if 0
        Trace("<<< %3d %p: read %5d/%ld %7d?\n",
            nf, cur, nr, (long)cur->size,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum);
#endif
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read;
            }

            Error("swapload: data, fail read() return %d %d (%d)",
                nr, c, errno);

            cur->len = 0;
            skcache_push_idle_frag(bk, cur);
            bk->frag_use--;
            cur = NULL;

            ret = 6;
            goto fail_read;
        }
        if(nr==0) {
            cur->len = 0;

            /* This frag is not first frag, remove that to save frags */
            if(nf>0) {
                skcache_push_idle_frag(bk, cur);
                bk->frag_use--;
                cur = NULL;
            }
            
            break;
        }

        cur->len = nr;
        nr_sum += nr;

        if(pre==NULL) {
			if(e->m_recorded) {
Trace("oid# %d, over write %p upon %p\n", e->oid, cur, e->m_recorded);
			}
            e->m_recorded = cur;
        }
        else {
            pre->next = cur;
        }

        nf++;
        pre = cur;
        cur->next = NULL;   /* dummy */
    } while(1);

    close(sfd);
    sfd = -1;

#if 0
    Trace("  swapload %ld bytes\n", (long)nr_sum);
#endif

    e->d_recorded = -1;


#if 0
    Trace("  swapload read %ld/%ld '%s'\n",
        (long)nr_sum, (long)e->totallength, name);
#endif


    cur = e->m_recorded;
#if 0
    if(cur)
    {
        int hlen;
        int xvmajor, xvminor;
        time_t xdate, xlmd, xexpire;
        off_t xclen;
        int xrcode;
        int xcachable;
        int xreload;

        hlen = HTTP_find_header_term(cur->ent, MIN(cur->len, e->totallength));

        chk = HTTP_parse_response_header(cur->ent, MIN(hlen, cur->len),
                &xvmajor, &xvminor, &xrcode,
                &xdate, &xlmd, &xexpire, &xclen,
                &xcachable, &xreload);

        if(chk) {
            Error("  *ignore* header");
        }

        Trace("  hlen %ld, xclen %ld + 1 -> %ld vs len %ld, diff %ld\n",
            (long)hlen, (long)xclen,
            (long)(xclen + hlen +1),
            (long)e->totallength,
            (long)(xclen + hlen +1 - e->totallength));

    }
#endif

    MUTEX_LOCK(&stat_lock);
    swapload_success++;
    MUTEX_UNLOCK(&stat_lock);

#if 0
    Trace("swapload: success p %p, oid# %d, bk# %d, %p\n", e, e->oid, bk->bid, bk);
#endif

    goto out;

fail_read_cntl:
    close(sfd);

#if 0
    Trace("swapload: fail read contorl block reason %d, erase %d, size %ld\n",
        ret, erase, (long)fsize);
#endif
    
    sim_stat_add_value(&swapload_fail_reason, ret);

    if(erase) {
        unlink(name);
        MUTEX_LOCK(&stat_lock);
        swap_count--;
        swap_size -= fsize;
        MUTEX_UNLOCK(&stat_lock);
    }

    goto out;

fail_read:
    close(sfd);

#if 0
    Trace("swapload: fail to read data. trush just readings, reason %d\n",
        ret);
#endif

    sim_stat_add_value(&swapload_fail_reason, ret);

    skcache_entry_purge(bk, e);

    goto out;

out:

#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapload_time, KOTETU_SUB_TIME(end, begin));

#if 0
    Trace("swapload: return %d\n", ret);
#endif

    return ret;
}




int
skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e)
{
    return skcache_entry_swapload(bk, e);
}









/*
 *
 * return
 *      0   success
 *      !0  fail
 *              >0  ignore
 *              <0  temporary, redo again
 */

int
skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;
    int chk;
    char *namechk;
    int i;
    struct timeval begin, end;
    struct stat sbuf;
    int sig_pos;
    int c;

    struct iovec iov[HARD_NFRAG_DWRITE];
    int     numf;
    off_t   lenf;

    off_t last_size;
    int lf;
    int sigpos;


    gettimeofday(&begin, NULL);

    ret = 0;

    sigpos = DSIG_POS(e->h1, e->h2, e->h3);


    if(bk==NULL || e==NULL) {
        Error("nil bucket pointer %x or entry pointer %x", bk, e);
        ret = 79;
        goto fail_start;
    }

#ifdef CHECK_MAGIC
    if(e->magic[0]!=MAGIC_0 || e->magic[1]!=MAGIC_1) {
        Error("swapout: strange magic %02x %02x / %02x %02x",
            e->magic[0], e->magic[1], MAGIC_0, MAGIC_1);

        ret = 4;
        goto fail_start;
    }
#endif

    if(e->m_recorded==NULL) {
        Trace("swapout: not recorded in memory bk# %d, e %p, length %ld\n",
            bk->bid, e, (long)e->totallength);
    }
#if 0
    if(e->m_recorded==NULL) {
        Error("swapout: not recorded in memory bk# %d, e %p, length %ld",
            bk->bid, e, (long)e->totallength);
        ret = 2;
        goto fail_start;
    }
#endif

#if 0
    if(bk->dirty2 == e->h2 && bk->dirty3 == e->h3) {
        Trace("swapout: conflict directory [%03d:%03d:%03d] to sweeper\n",
            e->h1, e->h2, e->h3);
        MUTEX_LOCK(&dir_mutex);
        dir_conflict++;
        MUTEX_UNLOCK(&dir_mutex);

        ret = 3;
        goto fail_start;
    }
#endif


    MUTEX_LOCK(&stat_lock);
    chk = swapout_token;
    if(chk>0) {
        swapout_token--;
        swapout_count++;
    }
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&stat_lock);

    if(chk<=0) {
        ret = -999;
        goto fail_start;
    }
    else
    if(chk==1) {
        Trace("swapout: *last* token of bk# %d, %p\n",
            bk->bid, bk);
    }

    if(lf>0) {
/*
        Trace("swapout: conflict directory bk# %d, dir# %d,.\n",
            bk->bid, sigpos);
*/
        MUTEX_LOCK(&dir_mutex);
        dir_conflict++;
        MUTEX_UNLOCK(&dir_mutex);

        ret = -777;
        goto fail_start;
    }

    cur = e->m_recorded;


    namechk = assign_filename(name, e);

    last_size = -1;

    c = 0;
try_open:
    c++;
    sfd = open(name, O_WRONLY | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(e->h1, e->h2, e->h3);
            goto try_open;
        }
        if(errno==ENOSPC) {
            Error("swapout: disk full on open()");
            MUTEX_LOCK(&stat_lock);
            swapout_token = 0;
            MUTEX_UNLOCK(&stat_lock);

            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("cannot open swapout file '%s' %d (%d)", name, c, errno);

        ret = 43;
        goto fail_open;
    }

try_lseek1:
    last_size = (off_t)lseek(sfd, 0, SEEK_END);
    if(last_size<(off_t)0L) {
        if(errno==EINTR) {
            goto try_lseek1;
        }
        Error("fail lseek() swap out file '%s' (%d)", name, errno);
        ret = 17;
        close(sfd);
        goto fail_open;
    }
    if(last_size!=(off_t)0) {
try_lseek2:
        chk = lseek(sfd, 0, SEEK_SET);
        if(chk<0 && errno==EINTR) {
            goto try_lseek2;
        }
    }



    /***
     *** write control
     ***/

    c = 0;
try_write_cntl:
    c++;
    nw = write(sfd, e, sizeof(skcache_entry_t));
    if(nw<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_write_cntl;
        }

        if(errno==ENOSPC) {
            Error("disk full!! at write control in swapout");
        }
        else
        if(errno==ENOMEM) {
            Error("kernel said no memory. reset token <%d>", c);

    MUTEX_LOCK(&stat_lock);
    swapout_token = 0;
    MUTEX_UNLOCK(&stat_lock);

            msleep(50);

            if(c<DISK_RETRY) {
                goto try_write_cntl;
            }

        }
        else {
            Error("fail write() return %d %d (%d)", nw, c, errno);
        }
        ret = 5;
        goto fail_write_cntl;
    }

    if(e->count>1) {
#if 0
    Trace("+# oid# %8d, %3d, %d, name %s\n",
        e->oid, e->count, e->hot, e->name);
#endif
    
        close(sfd);
        sfd = -1;
        goto skip_body_swapout;
    }
    if(cur==NULL) {
        close(sfd);
        sfd = -1;
        goto skip_body_swapout;
    }


    /***
     *** write body
     ***/

    nw_sum = 0L;
    i = 0;
    while(cur) {

        /* write frag chunk by writev() */
        numf = 0;
        lenf = (off_t)0L;
        while(cur && numf<nfrag_dwrite) {
#if 0
            Trace(">>> %3d %3d %p: stored %5d/%d %7ld\n",
                i, numf, cur->ent, cur->len, cur->size, (long)nw_sum+cur->len);
#endif
            if(cur->len>0) {
                iov[numf].iov_base  = cur->ent;
                iov[numf].iov_len   = cur->len;
                numf++;
                lenf += (off_t) cur->len;
            }
            cur = cur->next;
        }

        c = 0;
try_write:
        c++;
        nw = writev(sfd, iov, numf);
#if 0
        Trace(">>> %3d write %ld\n",
            i, nw);
#endif

        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write;
            }
            Error("fail write() return %d (%d)", nw, errno);
            if(errno==ENOSPC) {
                Error("disk full!! when swapout data");
            }

            ret = 7;
            goto fail_write;
        }
        else
        if(nw==0) {
            break;
        }
        if(nw>0) {
            nw_sum += nw;
        }

        i++;
    }
#if 0
    Trace("  name '%s'\n", name);
    Trace("  swap out %ld bytes '%s'\n", (long)nw_sum, name);
#endif

    ftruncate(sfd, (off_t) nw_sum + sizeof(skcache_entry_t) );
    close(sfd);
    sfd = -1;

skip_body_swapout:

    skcache_entry_purge(bk, e);
    

    e->m_recorded = NULL;
    e->d_recorded = e->oid;
#if 1
    e->swapout_candidate = 0;
#endif

#if 0
    Trace("swapout: oid# %d, write %ld, %ld+%ld/%ld '%s'\n",
        e->oid, sizeof(skcache_entry_t) + (long)nw_sum,
        sizeof(skcache_entry_t), (long)nw_sum, (long)e->totallength, name);
#endif

#if 0
    Trace("swapout: p %p, oid# %d, path '%s', size %ld\n", 
        e, e->oid, name, (long)nw_sum);
#endif

    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);


    MUTEX_LOCK(&stat_lock);
    swapout_success++;

    if(last_size<=0) {
        swap_count += 1LL;
        swap_size +=
            (long long)(nw_sum + (off_t)sizeof(skcache_entry_t));
    }
    else {
        swap_size +=
            (long long)(nw_sum + (off_t)sizeof(skcache_entry_t) - last_size);
    }
#if 0
    Trace("swapout: oid# %d, total swap count %lld size %lld\n",
        e->oid, swap_count, swap_size);
#endif
    MUTEX_UNLOCK(&stat_lock);


    MUTEX_LOCK(&dir_mutex);

    if(dir_count[sig_pos]<0) {
        /* nothing */
    }
    else {
        DSIG_SET(sig_pos, e->h4/DSIG_VALUE_RES);
        dir_sz[sig_pos] +=
            (off_t)(nw_sum + (off_t)sizeof(skcache_entry_t) - last_size);
        dir_count[sig_pos]++;
    }

    MUTEX_UNLOCK(&dir_mutex);

#if 0
    dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif

    goto out;

fail_write_cntl:
fail_write:
    close(sfd);

    unlink(name);

    sim_stat_add_value(&swapout_fail_reason, ret);

    MUTEX_LOCK(&stat_lock);
    if(last_size<0) {
    }
    else {
        swap_size -= last_size;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_open:
fail_start:

    sim_stat_add_value(&swapout_fail_reason, ret);

    goto out;

out:
#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapout_time, KOTETU_SUB_TIME(end, begin));

    return ret;
}





int
skcache_entry_purge_check_noref(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;

#if 0
    Trace("  entry_check_noref: p %p ref %d in bk%p ", p, p->ref, bk);
#endif

    if(p->ref>0) {
#if 0
        Trace("refered %d session, not purge\n", p->ref);
#endif
        ret = 0;
        goto skip;
    }
    else {
#if 0
        Trace("no refered , *purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }

skip:
    return ret;

}



int
skcache_entry_purge_check(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int erase;

    erase = 0;

    now = time(NULL);
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 0
        Trace("pc: empty issue_time\n");
#endif
        elap = skcache_mem_keep_max;
    }
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
        idle = skcache_mem_keep_cur/2;
    }

    if(p->ref>0) {
        erase = 0;
    }
    else
    if(p->expire>0 && p->expire >= now) {
        erase = 1;
    }
    else
    if(idle >= skcache_mem_keep_cur) {
        erase = 1;
    }

#if 0
    Trace("  check %p, ref %d, freq %d, idle %4d/%d, elap %5d: ",
        p, p->ref, p->count, idle, skcache_mem_keep_cur, elap);
#endif

    if(erase>0) {
#if 0
        Trace("*purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }
    else {
#if 0
        Trace("not purge\n");
#endif
        ret = 0;
    }

skip:
    return ret;

}


#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif



inline
static
int
_skcache_entry_swapout_check(void *pp, void *vv, int force)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int action;
    int chk;

    action = 0;

    now = time(NULL);

    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 1
        Error("_sw: empty issue_time entry %p", p);
#endif
        elap = -1;
    }

    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(skcache_mem_keep_cur/2, elap);
#endif
        idle = -1;
    }

#ifdef VS
    if(p->m_recorded) {
    Trace(" %p:: %8d %7ld %c 0x%08x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    Trace(" %p:: %8d %7ld %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif

#if 0
    Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */

    if(p->ref>0) {
#ifdef VS
Trace("refered ");
#endif
        action = 0;
    }
    else
    if(elap<0) {
#ifdef VS
Trace("junk    ");
#endif
        action = 2;
    }
    else
    if(p->cracked) {
#ifdef VS
Trace("cracked ");
#endif
        action = 2;
    }
    else
    if(p->d_recorded>=0) {
        if(p->m_recorded==NULL) {
#ifdef VS
Trace("already ");
#endif
            /* already swapouted */
            action = 0;
        }
        else {
#ifdef VS
Trace("mine-01 ");
#endif
            action = 2;
        }
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
Trace("empty-b ");
#endif
        action = 2;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
Trace("empty-c ");
#endif
        action = 2;
    }
    else
    if(elap > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire1 ");
#endif
        action = 2;
    }
    else
    if(p->expire>0 && p->expire <= now) {
#ifdef VS
Trace("expire2 ");
#endif
        action = 2;
    }
#if 0
    else
    if(p->lmd>0 && now - p->lmd > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire3 ");
#endif
        action = 2;
    }
#endif
    else {

  if(force) {

        if(idle < 0) {
#ifdef VS
Trace("neg,    ");
#endif
            action = 2;
        }
        else {
#ifdef VS
Trace("shift   ");
#endif
            action = 3;
        }

  }
  else {

#if 1
        /*
         * XXX
         * randomize forward swapout feature
         */
        if(idle<=skcache_mem_keep_min) {
            /* nothing */
        }
        else {
            idle += rand() % skcache_mem_keep_cur;
        }
#endif


        if(idle < 0) {
#ifdef VS
Trace("neg,    ");
#endif
            action = 2;
        }
        else
        if(idle >= skcache_mem_keep_cur) {
#ifdef VS
Trace("shift   ");
#endif
            action = 3;
        }
        else {
#ifdef VS
Trace("keep    ");
#endif
        }

  }


    }

    /* collect stat */
    if(action>=0 && action<4) {
        bk->op[action%4]++;
#ifdef VS
        Trace("op %d ", action);
#endif
    }

    ret = 0;
    switch(action) {
    case 3:
        chk = skcache_entry_swapout(bk, p);
        if(chk) {
            bk->inmemory++;
            ret = 0;
#ifdef VS
            Trace("*swapout-purge* fail (%d)\n", chk);
#endif

            if(chk==XSK_DISKFULL) {
                skcache_push_idle_entry(bk, p);
                p = NULL;
                ret = 1;
                break;
            }
        }
        else {
            bk->indisk++;
            skcache_push_idle_entry(bk, p);
            p = NULL;
            ret = 1;
#ifdef VS
            Trace("*swapout-purge* success\n");
#endif
        }
        break;

    case 2:
        chk = skcache_entry_purge(bk, p);
        if(chk) {
            bk->inmemory++;
            ret = 0;
#ifdef VS
            Trace("*purge* fail (%d)\n", chk);
#endif
        }
        else {
            skcache_push_idle_entry(bk, p);
            p = NULL;
            ret = 1;
#ifdef VS
            Trace("*purge*\n");
#endif
        }
        break;

    case 1:
        chk = skcache_entry_swapout(bk, p);
        if(chk) {
            bk->inmemory++;
#ifdef VS
            Trace("*swapout* fail (%d)\n", chk);
#endif
        }
        else {
            bk->inmemory++;
            bk->indisk++;
#ifdef VS
            Trace("*swapout* success\n");
#endif
        }

        break;

    default:
            bk->inmemory++;
#ifdef VS
        Trace("no-change\n");
#endif
        break;
    }

#if 0
    if(p) {
        if(p->m_recorded) {
            bk->inmemory++;
        }
        if(p->d_recorded>=0) {
            bk->indisk++;
        }
    }
#endif

skip:
    return ret;

}


int
skcache_entry_swapout_check(void *pp, void *vv)
{
    return _skcache_entry_swapout_check(pp, vv, 0);
}

int
skcache_entry_force_swapout_check(void *pp, void *vv)
{
    return _skcache_entry_swapout_check(pp, vv, 1);
}

int
skcache_dicision_w_th(bucket_t *bk, skcache_entry_t *p, int force,
    time_t now, time_t min, time_t th, time_t max)
{
    int elap, idle;
    int action;
    int chk;

#if 0
    Trace(
"_dicision_w_th: bk# %d, p 0x%x, force %d // min,thre,max %d,%d,%d\n",
        bk->bid, p, force, min, th, max);
#endif

    action = 0;

    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 1
        Error("_dc: empty issue_time");
#endif
        elap = -1;
    }

    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(th/2, elap);
#endif
        idle = -1;
    }

#ifdef VS
    if(p->m_recorded) {
    Trace(" %p:: %8d %7ld %c 0x%08x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    Trace(" %p:: %8d %7ld %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif

#if 0
    Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */

    if(p->ref>0) {
#ifdef VS
Trace("refered ");
#endif
        action = 0;
    }
    else
    if(elap<0) {
#ifdef VS
Trace("junk    ");
#endif
        action = 2;
    }
    else
    if(p->cracked) {
#ifdef VS
Trace("cracked ");
#endif
        action = 2;
    }
    else
    if(p->d_recorded>=0) {
        if(p->m_recorded==NULL) {
#ifdef VS
Trace("already ");
#endif
            /* already swapouted */
            action = 0;
        }
        else {
#ifdef VS
Trace("mine-01 ");
#endif
            /* XXX */
            action = 1;
        }
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
Trace("empty-b ");
#endif
        action = 2;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
Trace("empty-c ");
#endif
        action = 2;
    }
    else
    if(elap > max) {
#ifdef VS
Trace("expire1 ");
#endif
        action = 2;
    }
    else
    if(p->expire>0 && p->expire <= now) {
#ifdef VS
Trace("expire2 ");
#endif
        action = 2;
    }
#if 0
    else
    if(p->lmd>0 && now - p->lmd > max) {
#ifdef VS
Trace("expire3 ");
#endif
        action = 2;
    }
#endif
    else {

#if 0
        /*
         * XXX
         * randomize forward swapout feature
         */
        if(idle<=min) {
            /* nothing */
        }
        else {
            idle += rand() % th;
        }
#endif

        if(force) {
            if(idle < 0) {
#ifdef VS
    Trace("neg,    ");
#endif
                action = 2;
            }
            else {
#ifdef VS
    Trace("shift   ");
#endif
                action = 3;
            }
        }
        else {
            if(idle < 0) {
#ifdef VS
    Trace("neg,    ");
#endif
                action = 2;
            }
            else
            if(idle >= th) {
#ifdef VS
    Trace("shift   ");
#endif
                action = 3;
            }
            else {
#ifdef VS
    Trace("keep    ");
#endif
            }
        }
    }

skip:
    return action;

}

int
skcache_dicision(bucket_t *bk, skcache_entry_t *p, int force)
{
    time_t now;

    now = time(NULL);

    return skcache_dicision_w_th(bk, p, force,
        now,
        skcache_mem_keep_min, skcache_mem_keep_cur, skcache_entry_keep_cur);
}




extern pthread_mutex_t stat_lock;
extern volatile int compaction_count;

typedef struct {
    int no;
    int act;
    time_t date;
    off_t sz;
} tup;

int
comp_tup(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->act == y->act) {
        if(x->date == y->date) {
            return y->sz - x->sz;           /* y-x greater */
        }
        else {
            return x->date - y ->date;      /* x-y smaller */
        }
    }
    else {
        return y->act - x->act;             /* y-x greater */
    }
}

int
comp_tup_time(void*xp,void*yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    return x->date - y->date;
}



#ifndef TUP_MAX
/*
#define TUP_MAX (64)
*/
#define TUP_MAX (4096)
#endif

#define H4_DIR_COUNT_MAX    (128)
#define H4_DIR_SIZE_MAX     (128*K(10))
/*
#define H4_DIR_SIZE_MAX     (K(52))
*/

volatile int dir_count_max      = H4_DIR_COUNT_MAX;
volatile off_t dir_size_max     =  H4_DIR_SIZE_MAX;
volatile off_t dir_size_hwater  =  H4_DIR_SIZE_MAX;
volatile off_t dir_size_lwater  =  H4_DIR_SIZE_MAX/2;


/*
 * convert num-num -> num
 */
int name2num(char *nm)
{
    int r1, r2, r3, r4;
    int ret;
    register char *p, *q;
    register int i;
    char tmp[L_PATH];
    
    p = nm;

    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r3 = atoi(tmp);
    
    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r4 = atoi(tmp);

    ret = r3*radix_h4 + r4;

#if 0
    Trace("name2num '%s' -> %d\n", nm, ret);
#endif

    return ret;
}


int
rm_fr(char *base)
{
    char tmp[L_PATH];
    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    struct stat sbuf;
    int chk;

    if(base==NULL)
        return -1;

#if 0
Trace("rm_fr: called '%s'\n", base);
#endif

    chk = rmdir(base);
    if(chk) {
        if(errno==ENOTEMPTY) {

            dirp = opendir(base);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.' && dp->d_name[1]=='\0') {
                        continue;
                    }
                    if(dp->d_name[0]=='.'
                        && dp->d_name[1]=='.' && dp->d_name[2]=='\0') {
                        continue;
                    }
                    /*
                    strcpy(tmp, base);
                    strcat(tmp, "/");
                    strcat(tmp, dp->d_name);
                    */
                    pathncat(tmp, L_PATH, base, dp->d_name);


                    chk = stat(tmp, &sbuf);
                    if(chk) {
                        Error("rm_fr: fail stat() '%s' (%d)",
                            tmp, errno);
                        continue;
                    }

                    if(S_ISREG(sbuf.st_mode)) {
                        unlink(tmp);
                    }
                    if(S_ISDIR(sbuf.st_mode)) {
                        rm_fr(tmp);
                    }
                }
                closedir(dirp);
            }

            chk = rmdir(base);

            if(chk) {
                Error("rm_fr: fail rmdir() '%s' after sweep (%d)",
                    base, errno);
            }

        }
    }

    return 0;
}

int
skcache_check_dir(int d1, int d2, int d3, int force)
{
    char basename[L_PATH];
    char dirname[L_PATH];
    char tmp[L_PATH];
    char *namechk;
    dirent_t    *elem;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    char name[L_PATH];
    int dfd;
    char buf[L_BUF];
    int n;
    int i;
    int nr;
    int chk;
    off_t sum;
    off_t fsize;
    int lcount, dcount, ecount;
    off_t lsize, dsize, esize;
    int strange = 0;
    int action;
    int life;
    time_t reftime;
    time_t now;
    int nfile;

    tup rec[TUP_MAX];

    struct timeval begin, end;

    bucket_t *bk;
    skcache_entry_t en;
    int c;
    int er=0;

    int igh=0;

/*
    int sig = 0L;
    int64_t sig = 0LL;
*/
    u_int32_t sig[DSIG_ELEM_WIDTH];

    int sig_pos = 0;

    gettimeofday(&begin, NULL);
    reftime = begin.tv_sec;


#if 0
Trace("check_dir: d1,d2,d3 %4d,%4d,%4d time %d force? %d\n",
   d1, d2, d3, reftime, force);
#endif

    if(d1<0||d1>=RADIX_H1_MAX
        || d2<0||d2>=RADIX_H2_MAX
            || d3<0||d3>=RADIX_H3_MAX) {
        Error("check_dir: ignore directory position %d,%d,%d", d1, d2, d3);
        goto end;
    }

    if(d1>=radix_h1 || d2>=radix_h2 || d3>=radix_h3) {
        if(force) {
#if 1
            Trace("check_dir: out of current directory range <%d,%d,%d>, try clear\n",
                d1, d2, d3);
#endif

#if 1
            namechk = assign_dirname_bynumber(dirname, d1, d2, -1);
            if(namechk) {
#if 1
                Trace("check_dir: clear dir. recursive %s\n", dirname);
#endif
                rm_fr(dirname);
            }
#endif
#if 0
            namechk = assign_dirname_bynumber(dirname, d1, d2, d3);
            if(namechk) {
                rm_fr(dirname);
            }
#endif
        }
        else {
#if 1
            Trace("check_dir: out of current directory range <%d,%d,%d>, no change\n",
                d1, d2, d3);
#endif
        }

        goto end;
    }



    now = time(NULL);
    sig_pos = DSIG_POS(d1, d2, d3);

    dsig_allclear(sig);

#if 0
    Trace("-=-=-=\n");
    dsig_show(sig);
#endif


#if 0
Trace("check_dir: dir# %d, %d,%d,%d time %d force? %d\n",
    sig_pos, d1, d2, d3, reftime, force);
#endif

    bk = &skcache_bucket_base[d1];

#ifdef VD3
Trace("check_dir: %p dir# %d, [%03d:%03d:%03d] th %08x %d %s\n",
  bk, sig_pos, d1, d2, d3, pthread_self(), reftime, force ? "FORCE" : "normal");
#endif
#if 0
Trace("    %4d/%4d/%4d\n", radix_h1, radix_h2, radix_h3);
#endif

#if 0
    /* obsolted by dir_lock[] */
    LOCK_BK(bk);
 assert(bk->cup.num==bk->entry_use);

    if(bk->dirty2 >= 0 || bk->dirty3 >= 0) {
        Error(
"check_dir: %p [%03d:%03d:%03d], [%03d:%03d:%03d] is used by another thread",
            bk, d1, d2, d3, d1, bk->dirty2, bk->dirty3);
        UNLOCK_BK(bk);
        goto out;
    }

    bk->dirty2 = d2;
    bk->dirty3 = d3;

    UNLOCK_BK(bk);
#endif


disk_operations:


    n = 0;
    sum = (off_t)0L;

    namechk = assign_dirname_bynumber(dirname, d1, d2, d3);

#ifdef VD3
    Trace("  target dir# %d, path '%s'\n", sig_pos, dirname);
#endif


    /***
     *** scan directories
     ***/

    c = 0;
try_opendir:
    c++;
    dirp = opendir(dirname);
    if(dirp==NULL) {
        if(errno==ENOENT && c<DISK_RETRY) {
            if(force) {
#ifdef VD
                Trace("check_dir: not found dir# %d, '%s', build that.\n",
                    sig_pos, dirname);
#endif
                build_dir_bynumber(d1, d2, d3);

                MUTEX_LOCK(&dir_mutex);
                checkdir_dir_count++;

#if 0
                dir_sig[sig_pos]    = 0L;
#endif
                DSIG_ALLCLEAR(sig_pos);
                dir_sz[sig_pos]     = (off_t)0L;
                dir_count[sig_pos]  = 0L;

                MUTEX_UNLOCK(&dir_mutex);

                msleep(10);     /* XXX */
            }

            goto fail_opendir;
        }
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_opendir;
        }

        Error("fail opendir() '%s' %d (%d)", dirname, c, errno);

        goto fail_opendir;
    }

    i = 0;
    while(1) {
        chk = readdir_r(dirp, ent, &elem);
        if(chk<0) {
            if(errno==EINTR) {
                continue;
            }
            Error("fail readdir_r() for '%s' (%d)", dirname, errno);
            break;
        }
        if(elem==NULL) {
#if 0
            Trace("dir '%s' is empty\n", dirname);
#endif
            break;
        }

        if(elem->d_name[0]=='.' && elem->d_name[1]=='\0') {
            continue;
        }
        if(elem->d_name[0]=='.'
            && elem->d_name[1]=='.' && elem->d_name[2]=='\0') {
            continue;
        }

        i++;

        /*
        strcpy(name, dirname);
        strcat(name, "/");
        strcat(name, elem->d_name);
        */

        pathncat(name, L_PATH, dirname, elem->d_name);

#ifdef VD
        Trace("    %6d %4d : %-14s %s\n", sig_pos, i, elem->d_name, name);
#endif


        life = -1;
        action = 0;

        c = 0;
try_open:
        c++;
        dfd = open(name, O_RDONLY);
        if(dfd<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_open;
            }
            Error("check_dir: fail open() '%s' %d (%d)",
                name, c, errno);
            action = 1;
            goto next;
        }

        /* count file size by seek end
            and seek first to read control block */
        c = 0;
try_lseek:
        c++;
        fsize = lseek(dfd, 0, SEEK_END);
        if(fsize<=(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_lseek;
            }
            Error("check_dir: fail lseek() '%s' %d (%d)",
                name, c, errno);
            action = 1;
            goto done_read_cntl;
        }
        if(fsize>(off_t)0) {
            lseek(dfd, 0, SEEK_SET);
        }

        sum += fsize;

        c = 0;
try_read_cntl:
        c++;
        nr = read(dfd, &en, sizeof(skcache_entry_t));
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read_cntl;
            }
            Error("check_dir: file# %d, fail read() '%s' %d (%d)",
                n, name, c, errno);
            action = 1;
            goto done_read_cntl;
        }

        if(nr<sizeof(skcache_entry_t)) {
            action = 2;
#if 1
            Error("check_dir: ignore file# %d, too short %d/%d",
                n, nr, sizeof(skcache_entry_t));
#endif
            en.issue_time = -1;
            goto done_read_cntl;
        }

        if(en.magic[0]!=MAGIC_0 || en.magic[1]!=MAGIC_1) {
            action = 2;
#ifdef CHECK_MAGIC
 Error("check_dir: ignore file# %d, wrong magic %02x %02x / %02x %02x '%s'",
        n, en.magic[0], en.magic[1], MAGIC_0, MAGIC_1, name);
#endif
            en.issue_time = -1;
            goto done_read_cntl;
        }

 {
    int xh1, xh2, xh3, xh4;

        calc_hash(en.name, &xh1, &xh2, &xh3, &xh4);

        if(en.h1!=xh1 || en.h2!=xh2 || en.h3!=xh3 || en.h4!=xh4) {
            action = 2;
            igh++;
#if 0
Trace("\
check_dir: ignore hash, file '%s'\n\
    stored %03d:%03d:%03d:%03d %s\n\
    solved %03d:%03d:%03d:%03d same\n",
    name,
    en.h1, en.h2, en.h3, en.h4, en.name,
    xh1, xh2, xh3, xh4);
#endif
#if 0
Trace("check_dir: ignore hash, stored %03d:%03d:%03d:%03d, we guess %03d:%03d:%03d:%03d\n",
        en.h1, en.h2, en.h3, en.h4, xh1, xh2, xh3, xh4);
#endif
            en.issue_time = -1;
            goto done_read_cntl;
        }

 }

        if(en.issue_time<=0) {
            action = 1;
        }
        else {
#if 1
            /*XXX*/
            if(en.issue_time<min_date) {
                action = 2;
            }
            else
            if(en.issue_time>reftime) {
                action = 2;
            }
            else
#endif
            if(reftime - en.issue_time > skcache_entry_keep_cur) {
                action = 1;
            }
        }

        if(!action) {
            if(en.lmd<=0) {
                action = 1;
            }
#if 0
            if(en.lmd>0 && reftime - en.lmd > skcache_entry_keep_cur) {
                action = 1;
            }
#endif

            if(en.expire>0 && reftime - en.expire > 0) {
                action = 1;
            }
        }


done_read_cntl:
        (void) 1;

read_done:

next:

        close(dfd);

#ifdef VD2
        Trace(" f %4d %4d %4d %d %d %8d %9ld %s\n",
            i, n, atoi(elem->d_name), action,
            en.issue_time, reftime - en.issue_time, (long)fsize, name);
#endif

        if(n>=TUP_MAX) {
            er++;
            if(er==1) {
                Error("dir '%s', tup overflow read %d th, store %d th in %d",
                    dirname, i, n, TUP_MAX);
            }
            unlink(elem->d_name);
        }
        else {
#ifdef USE_3LEVEL_DIR
            rec[n].no = atoi(elem->d_name);
#else
            rec[n].no = name2num(elem->d_name);
#endif

/*
Trace("n %d %d <- %s\n", n, rec[n].no, elem->d_name);
*/

            rec[n].act = action;
            rec[n].date = en.issue_time;
            rec[n].sz = fsize;
            n++;
        }

#if 0
        if(i%256==255) {
            msleep(1);
        }
#endif
    }

    closedir(dirp);

    nfile = i;

#ifdef VD
    Trace("  dir# %d, nfile %d, record %d/%d, \n",
        sig_pos, nfile, n, TUP_MAX);
#endif


    /***
     *** main body of compaction ***/

    qsort(rec, n, sizeof(tup), comp_tup);

    lcount  = 0;
    dcount  = 0;
    lsize   = 0;
    dsize   = 0;
    ecount  = 0;
    esize   = 0;

#if 0
    if(dir_sz[sig_pos]>=dir_size_hwater) {
    Trace("  check_dir: dir# %d, read sum %lld, hwater %lld, lwater %lld\n",
        sig_pos,
        (long long)sum,
        (long long)dir_size_hwater, (long long)dir_size_lwater);
    }
#endif

    for(i=0;i<n;i++) {
        int xh1, xh2, xh3, xh4;

#ifndef USE_3LEVEL_DIR
        xh3 = rec[i].no/radix_h4;
        xh4 = rec[i].no%radix_h4;

        sprintf(name, "%s/%d-%d", dirname, xh3, xh4);
#else
        xh4 = rec[i].no;
        sprintf(name, "%s/%d", dirname, xh4);
#endif

/*
Trace("i %d no %d xh4 %d -> %s\n", i, rec[i].no, xh4, name);
*/


#if 0
        Trace("%4d: no# %7d act %2d date %d size %9ld\n",
            i, rec[i].no, rec[i].act, rec[i].date, rec[i].sz);
#endif

        if(rec[i].act==0) {
            if(sum - esize - dsize > dir_size_lwater) {
#if 0
                Trace("\t\tremove 1\n");
#endif
                rec[i].act = 1;
            }
            else {
                if(rec[i].sz > dir_size_lwater) {
#if 0
                    Trace("\t\tremove 2\n");
#endif
                    rec[i].act = 1;
                }
                else {
#if 0
                    Trace("\t\tkeep\n");
#endif
                }
            }
        }

        switch(rec[i].act) {
        case 0:
            /* mark with xh3 and xh4 */
            DSIG_TMP_SET(sig, xh4/DSIG_VALUE_RES);

#if 0
            Trace("  xh4 %d, v %d (%d), byte %d/%d, bit %d\n",
                xh4, xh4/DSIG_VALUE_RES, DSIG_VALUE_RES,
                (xh4/DSIG_VALUE_RES)/(SIZE_OF_UINT*7),
                DSIG_ELEM_WIDTH,
                (xh4/DSIG_VALUE_RES)%(SIZE_OF_UINT*8));
            dsig_show(sig);
#endif

            lcount++;
            lsize += rec[i].sz;
            break;

        case 1:

            c = 0;
try_unlink1:
            c++;
            chk = unlink(name);
            if(chk) {
                if(errno==EINTR) {
                    goto try_unlink1;
                }
                if(errno==EISDIR) {
                    rm_fr(name);
                }
                else {
                    Error("check_dir: 1 - fail unlink() %s %d (%d)",
                        name, c, errno);
                }
            }

            dcount++;
            dsize += rec[i].sz;
            break;

        case 2:
        default:

            c = 0;
try_unlink2:
            c++;
            chk = unlink(name);
            if(chk) {
                if(errno==EINTR) {
                    goto try_unlink2;
                }
                Error("check_dir: 2 - fail unlink() %s %d (%d)",
                    name, c, errno);
            }

            ecount++;
            esize += rec[i].sz;
            break;

        }

#if 0
        Trace("\tl %9ld d %9ld e %9ld rest %9ld\n",
            lsize, dsize, esize, sum - esize - dsize);
#endif

    }

    now = time(NULL);

#ifdef VD
    Trace(" dd %6d %-24s %4d %8x %7ld + %d\
/ live %4d %7ld / dead %4d %7ld\
/ old %4d %7ld / %2d\n",
        sig_pos, dirname, n, sig, (long)sum, strange,
        lcount, (long)lsize, dcount, (long)dsize, ecount, (long)esize,
        now - reftime
        );
#endif

    MUTEX_LOCK(&stat_lock);

    checkdir_dir_count++;
    checkdir_file_read_count    += nfile;
    checkdir_file_handle_count  += n;
    checkdir_file_left_count    += lcount;
    checkdir_file_delete_count  += dcount;
    checkdir_file_erase_count   += ecount;

    check_dir_ignore_hash_count = igh;

    if(dcount>0 && dsize>0) {
        swap_count  -= dcount;
        swap_size   -= dsize;

#if 0
        if(swap_count<=0LL || swap_size<=0LL) {
            Trace("swap count or size is negative, count %lld size %lld\n",
                swap_count, swap_size);
        }
#endif
    }

    MUTEX_UNLOCK(&stat_lock);

    if(force) {
        MUTEX_LOCK(&dir_mutex);

#if 0
        dir_sig[sig_pos]    = sig;
#endif
        memcpy((void*)&dir_sig[sig_pos*DSIG_ELEM_WIDTH],
            (void*)&sig, DSIG_ELEM_WIDTH*sizeof(u_int32_t));

        dir_sz[sig_pos]     = lsize;
        dir_count[sig_pos]  = lcount;
        dir_time[sig_pos]   = now;

        MUTEX_UNLOCK(&dir_mutex);
    }

#if 0
    dsig_show(sig);
    dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif


fail_opendir:

#if 0
    if(bk) {
        LOCK_BK(bk);

        bk->dirty2 = -1;
        bk->dirty3 = -1;

        UNLOCK_BK(bk);
    }
#endif

out:


end:
    gettimeofday(&end, NULL);
    sim_stat_add_value(&checkdir_time, KOTETU_SUB_TIME(end, begin));
    sim_stat_add_value(&checkdir_time_short, KOTETU_SUB_TIME(end, begin));

#ifdef TRACE_CHKDIR
    Trace("  done chkdir dir# %d, %-17s left %4df %8ldB %7.2fKB  th %08x\n",
            sig_pos, dirname, lcount, (long)lsize, fdK(lsize), pthread_self());
#endif

    return 0;
}


volatile int dsweeper_seq   = 0;
volatile int dsweeper_count = 0;
volatile int dsweeper_phase = -1;
volatile int dsweeper_pos   = -1;
volatile int dsweeper_full_mode = 0;

/* how many tics are required to work sweeper */
#ifndef SWEEP_CYCLE
#define SWEEP_CYCLE (1)
#endif
#ifndef SWEEP_BOOT_CYCLE
#define SWEEP_BOOT_CYCLE (1)
#endif

/* how many direcotry are sweeped per chance */
#ifndef SWEEP_NTRY
#define SWEEP_NTRY  (1)
#endif


volatile int dsweep_boot_cycle = SWEEP_BOOT_CYCLE;
volatile int dsweep_cycle = SWEEP_CYCLE;



/* for two level */
int
skcache_dir_sweeper(void)
{
    int f;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
    int nd1;
    int pos;

    MUTEX_LOCK(&stat_lock);
    if(dsweeper_phase<0) {
        dsweeper_phase = 0;
        dsweeper_pos = rand() % dir_num;
    }
    p = dsweeper_phase;
    f = dsweeper_seq;
    dsweeper_seq =  (dsweeper_seq+1) % 1000000;

#if 0
    if(dsweeper_count%dir_num==0) {
        Trace("first dir\n");
    }
#endif
    dsweeper_count++;

    v = dsweeper_pos;

    if(dsweeper_pos>=dir_num-1) {
        dsweeper_phase++;
        dsweeper_pos = 0;
    }
    else {
        dsweeper_pos++;
    }
    MUTEX_UNLOCK(&stat_lock);

    if(f<0) {
        f = v;
    }

    d1  = v%radix_h1;
    d2  = v/radix_h1;
    d3  = 0;

    pos = DSIG_POS(d1, d2, d3);

#if 0
    Trace("  sweeper th %08x %6d %6d -> dir# %d, %03d %03d %03d\n",
        pthread_self(), p, v, pos, d1, d2, d3);
#endif
    MUTEX_LOCK(&dir_mutex);
    if(dir_lock[pos]>0) {
        pos = -1;
    }
    MUTEX_UNLOCK(&dir_mutex);

    if(pos>0) {
        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = pthread_self();
        MUTEX_UNLOCK(&dir_mutex);

        chk = skcache_check_dir(d1, d2, d3, 1);

        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = 0;
        MUTEX_UNLOCK(&dir_mutex);
    }

done:

    return 0;
}

int
skcache_dir_sweeper_level3(void)
{
    int f, i;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
    int nd1;

    int dir_max=RADIX_H1_MAX*RADIX_H2_MAX*RADIX_H3_MAX;

    MUTEX_LOCK(&stat_lock);
    p = dsweeper_phase;
    f = dsweeper_seq;
    dsweeper_seq =  (dsweeper_seq+1) % 1000000;
    MUTEX_UNLOCK(&stat_lock);

    if(p==1) {
        goto phase_one;
    }

phase_zero:
    if(f%dsweep_boot_cycle!=0) {
        return 0;
    }

#if 0
    Trace("sweep 0: th %08x f %9d -> %4d\n",
        pthread_self(), f, f%dsweep_boot_cycle);
#endif

    MUTEX_LOCK(&stat_lock);
    dsweeper_count++;
    v = dsweeper_pos;

    if(dsweeper_full_mode) {
        dsweeper_pos = (dsweeper_pos + 1 ) % dir_max;

        d3 = (v/(RADIX_H2_MAX*RADIX_H1_MAX))%RADIX_H3_MAX;
        d2 = (v%(RADIX_H2_MAX*RADIX_H1_MAX))/RADIX_H1_MAX;
        d1 = (v%(RADIX_H2_MAX*RADIX_H1_MAX))%RADIX_H1_MAX;

    }
    else {
        dsweeper_pos = (dsweeper_pos + 1 ) % dir_num;

        d3 = (v/(radix_h2*radix_h1))%radix_h3;
        d2 = (v%(radix_h2*radix_h1))/radix_h1;
        d1 = (v%(radix_h2*radix_h1))%radix_h1;
    }
    MUTEX_UNLOCK(&stat_lock);

#if 0
    Trace("  p0 th %08x %8d -- %6d -> [%03d:%03d:%03d] %s\n",
        pthread_self(), f, v, d1, d2, d3,
        dsweeper_full_mode ? "FULL" : "normal");
#endif

    if(dsweeper_full_mode) {
        char dirname[L_PATH];
        char *namechk;

        if(d1>=radix_h1 || d2>=radix_h2 || d3>=radix_h3) {
            namechk = assign_dirname_bynumber(dirname, d1, d2, -1);
            if(namechk) {
#if 1
                Trace("sweeper: clear dir %03d:%03d:%03d %s\n",
                    d1, d2, d3, dirname);
#endif
                rm_fr(dirname);
            }
        }
        else {
            chk = skcache_check_dir(d1, d2, d3, 0);
        }
    }
    else {
        chk = skcache_check_dir(d1, d2, d3, 1);
    }


    MUTEX_LOCK(&stat_lock);
    if(dsweeper_pos == 0) {
        dsweeper_phase = 1;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto done;

phase_one:
    if(f%dsweep_cycle!=0) {
        return 0;
    }
#if 0
    Trace("sweep 1: th %08x f %9d -> %4d\n",
        pthread_self(), f, f%dsweep_cycle);
#endif

    for(i=0;i<SWEEP_NTRY;i++) {
        MUTEX_LOCK(&stat_lock);
        if(dsweeper_count%dir_num==0) {
            Trace("first dir\n");
        }
        dsweeper_count++;
        v = dsweeper_pos;
        dsweeper_pos = (dsweeper_pos + 1 ) % dir_num;
        MUTEX_UNLOCK(&stat_lock);

        if(f<0) {
            f = v;
        }
        d3 = (v/(radix_h2*radix_h1))%radix_h3;
        d2 = (v%(radix_h2*radix_h1))/radix_h1;
        d1 = (v%(radix_h2*radix_h1))%radix_h1;
#if 0
        Trace("  p1 th %08x %8d %2d %6d -> [%03d:%03d:%03d]\n",
            pthread_self(), f, i, v, d1, d2, d3);
#endif
        chk = skcache_check_dir(d1, d2, d3, 1);

    }

done:

    return 0;
}





int
skcache_calc_param(long long want,
    int *xbk_num, int *xsk_level, int *xffl_size, double r,
    int *xsk_len, int *xffl_num)
{
    long long bksize;
    long a, b, c;

#if 0
    Trace("skcache_calc_param: want %lld\n", want);
#endif

#if 0
    Trace("    specified xbk_num %d, xsk_level %d, xffl_size %d r %.2f\n",
        *xbk_num, *xsk_level, *xffl_size, r);
#endif

    if(*xffl_size<0) {
        *xffl_size = 1460;
    }

    bksize = want/(*xbk_num) - sizeof(skcache_entry_t);

    a = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    b = sizeof(skcache_entry_t);
    c = sizeof(skcache_frag_t)+ *xffl_size;

#if 0
    Trace("    bucket size %lld\n", bksize);
    Trace("    a %ld\n", a);
    Trace("    b %ld\n", b);
    Trace("    c %ld\n", c);
#endif

    *xsk_len = (long)
        (bksize - b) / (a + b + (long)(r*c));
    *xffl_num = (long)(r * *xsk_len);

#if 0
    Trace("    solved    xsk_len %d, xffl_num %d\n",
        *xsk_len, *xffl_num);
#endif


    return *xsk_len;
}


long long
skcache_calc_size(
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_len, int *xffl_size)
{
    long long r;
    int esk;
    int tsk;
    int ten;
    int tfl;
    long long ebk;
    long long tbk;

#if 0
    Trace("skcache_calc_size:\n");
    Trace("  bucket:\n");
#endif

    r = (((sizeof(void*)+sizeof(skiplist_cell_t**)+sizeof(skiplist_cell_t*)*(*xsk_level+1)) + sizeof(skcache_entry_t))* (*xsk_len)
        + sizeof(skcache_entry_t)
        + (sizeof(skcache_frag_t) + *xffl_size) * (*xffl_len))*(*xbk_num);

    esk = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    tsk = esk * *xsk_len;
#if 0
    Trace("    skiplist:\n");
    Trace("      level %d\n", *xsk_level);
    Trace("      -> element size of skiplist %ld\n", esk);
    Trace("      length of skiplist %d\n", *xsk_len);
#endif

    ten = sizeof(skcache_entry_t)*(*xsk_len+1);
    tfl = (sizeof(skcache_frag_t)+ *xffl_size)*(*xffl_len);
    ebk = (long long)(sizeof(bucket_t)+tsk+ten+tfl);
#if 0
    Trace("    -> size of skiplist            %12ld %4ldM+%07ld %7.2f%%\n",
        tsk, (long)dM(tsk), (long)mM(tsk), ((double)tsk*100)/ebk);
    Trace("    size of free-list for entities %12ld %4ldM+%07ld %7.2f%%\n",
        ten, (long)dM(ten), (long)mM(ten), ((double)ten*100)/ebk);
    Trace("    size of fragment free-list     %12ld %4ldM+%07ld %7.2f%%\n",
        tfl, (long)dM(tfl), (long)mM(tfl), ((double)tfl*100)/ebk);
    Trace("  -> size of bucket                %12lld %4ldM+%07ld\n",
        ebk, (long)dM(ebk), (long)mM(ebk));
#endif


    tbk = ebk * *xbk_num;

#if 0
    Trace("  number of bucket %ld\n", *xbk_num);
    Trace("-> total size of all buckets       %12lld %4ldM+%07ld\n",
        tbk, (long)dM(tbk), (long)mM(tbk));
    Trace("-> total number of entries         %12ld\n", *xsk_len * *xbk_num);
    Trace("-> total number of frags           %12ld\n", *xffl_len * *xbk_num);
#endif

    return r;
}




int check_bkinfo();


skcache_entry_t*
skcache_pull_new_entry(bucket_t *bk)
{
    int chk;
    skcache_entry_t* newone;

#if 0
#ifndef NDEBUG
    Trace("new_entry: bk# %d, %p, use %d, free %d, cup.num %d\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num);
#endif
#endif

    if(bk->entry_freelist == NULL) { 
        Error("new_entry: no free item");
        return NULL;
    }

    newone = bk->entry_freelist;
    bk->entry_freelist = newone->next;
    bk->entry_use++;
    bk->entry_free--;

    memset(newone, 0, sizeof(skcache_entry_t));
    newone->magic[0] = MAGIC_0;
    newone->magic[1] = MAGIC_1;
    newone->h1 = newone->h2 = newone->h3 = newone->h4 = -1;
    newone->ref = 0;
    newone->issue_time = -1;
    newone->swapout_candidate = 0;

    return newone;
}

int
skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen)
{
#if 0
#ifndef NDEBUG
    Trace("idle_entry: bk# %d, %p, use %d, free %d, cup.num %d, target %p\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num, xen);
#endif
#endif

    memset(xen, 0, sizeof(skcache_entry_t));
    xen->h1 = xen->h2 = xen->h3 = xen->h4 = -1;
    xen->ref = -1;
    xen->issue_time = -1;
    xen->swapout_candidate = -1;

    xen->next = bk->entry_freelist;
    bk->entry_freelist = xen;
    bk->entry_use--;
    bk->entry_free++;

    return 0;
}



skcache_frag_t*
skcache_pull_new_frag(bucket_t *bk)
{
    skcache_frag_t* cur;
    int chk;
    int c;

#if 0
#ifndef NDEBUG
    Trace("new_frag: use %d, free %d\n",
        frag_use, frag_free);
#endif
#endif

    c = FRAG_NTRY;
try_lock:
    c--;
    chk = MUTEX_TRYLOCK(&frag_lock);
    if(chk==EBUSY) {
        if(c>0) {
#if 0
            Trace("new_frag: free list is busy, retry %d, u %d, f %d\n",
                c, frag_use, frag_free); 
#endif
            msleep(3);
            goto try_lock;
        }
        Error("new_frag: free list is busy <%d times>, u %d, f %d",
            c, frag_use, frag_free);
        return NULL;
    }

    if(z_frag_freelist == NULL) { 
        Error("new_frag: no free item, use %d %.2f%%",
            frag_use, (double)100.0*frag_use/(frag_use+frag_free));

        MUTEX_UNLOCK(&frag_lock);
        return NULL;
    }

    cur = z_frag_freelist;
    z_frag_freelist = cur->next;
    frag_use++;
    frag_free--;
    cur->next = NULL;
    cur->len  = 0;

    MUTEX_UNLOCK(&frag_lock);

    return cur;
}

int
RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    xf->next = z_frag_freelist;
    z_frag_freelist = xf;
    frag_use--;
    frag_free++;

    return 0;
}

int
skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    int chk;
    int c;

#if 0
#ifndef NDEBUG
    Trace("idle_frag: target %p, use %d, free %d\n",
        xf, frag_use, frag_free);
#endif
#endif

    c = FRAG_NTRY*2;
try_lock:
    c--;
    chk = MUTEX_TRYLOCK(&frag_lock);
    if(chk==EBUSY) {
        if(c>0) {
#if 0
            Trace("idle_frag: free list is busy, try %d, u %d, f %d\n",
                c, frag_use, frag_free);
#endif
            msleep(5);
            goto try_lock;
        }
        Error("idle_frag: free list is busy, give up, u %d, f %d",
            frag_use, frag_free);
        return -1;
    }

#if 0
    xf->next = z_frag_freelist;
    z_frag_freelist = xf;
    frag_use--;
    frag_free++;
#endif
    chk = RAW_skcache_push_idle_frag(bk, xf);

    MUTEX_UNLOCK(&frag_lock);

    return 0;
}


#if 0
int
check_checkpoints()
{
    int i;
    time_t now;
    bucket_t *bk;
    int width = skcache_mem_keep_min * 3;

    now = time(NULL);

    for(i=0;i<radix_h1;i++) {
        bucket_t *bk;

        bk = &skcache_bucket_base[i];

        LOCK_BK(bk);

        if(bk->entry_use >= bk->entry_lwater
            && now - bk->gc_last > width )
        {
            int chk;

    MUTEX_LOCK(&stat_lock);
    compaction_count++;
    MUTEX_UNLOCK(&stat_lock);
            chk = mark_target(bk, XSK_ENTRY);

            bk->gc_last = now + (rand() % width) + i ;  /* XXX */
        }

        UNLOCK_BK(bk);
    }

}
#endif


/*
 * re-calcurate keeptime threshould
 * argument:
 *      delta_t [msec]
 */
int
skcache_update_keeptime(int delta_t)
{
    time_t now;
    int last, keep;
    double ratio;
    int rapid_flag;
    int i;
    bucket_t *bk;
    int chk;
    char *msg;

    now = time(NULL);
    rapid_flag = 0;

    MUTEX_LOCK(&config_lock);
    ratio = sim_hist_average_d(&fratio_hist);
    last = keep = skcache_mem_keep_cur;

    if(skcache_keeptime_token>0) {
        skcache_keeptime_token--;
        goto skip_update;
    }

    /* slow up and rapid down */
    if(ratio>skcache_frag_hwater_ratio) {
        msg = "rapid-down";
        skcache_mem_keep_cur = 9*keep/10;
        skcache_keeptime_token = skcache_keeptime_token_init;
    }
    else
    if(ratio>skcache_frag_lwater_ratio) {
        msg = "no-change";
    }
    else {
        msg = "up";
        skcache_mem_keep_cur = keep + delta_t/1000 + 1;
    }

    if(skcache_mem_keep_cur<skcache_mem_keep_min) {
        Trace("  underflow min %d\n", skcache_mem_keep_min);
        skcache_mem_keep_cur = skcache_mem_keep_min;
    }
    if(skcache_mem_keep_cur>skcache_mem_keep_max) {
        Trace("  overflow  max %d\n", skcache_mem_keep_max);
        skcache_mem_keep_cur = skcache_mem_keep_max;
    }

    Trace("KT updating %d frag %.2f%% (%d) %s thre. %d -> %d\n",
        now, ratio, skcache_keeptime_token, msg, last, skcache_mem_keep_cur);

skip_update:
    MUTEX_UNLOCK(&config_lock);

    if(rapid_flag) {
        for(i=0;i<radix_h1;i++) {
            bk = &skcache_bucket_base[i];

            chk = MUTEX_TRYLOCK(&bk->mutex);
            if(chk==EBUSY) {
            }
            else {
                bk->gc_push++;
                MUTEX_UNLOCK(&bk->mutex);
            }

        }
    }
}

/*
 * When it find @ (at-mark) in prefix of cache dirctory,
 * the program recognize using many disk and its pattern.
 */
check_disk_spliting()
{
    char *p;
    int v, i;
    char tmp[L_PATH];

    v = -1;
    p = cache_dir_prefix;
    while(*p) {
        if(*p=='@') {
            v = atoi(p+1);
            *p = '\0';
            break;
        }
        p++;
    }

    if(v>0) {
        radix_disk = v;

#if 0
        Trace("cache directories are splited to %d disks\n", radix_disk);
#endif
        Trace(" Spricing: cache directories stored into %d disks\n",
            radix_disk);
        Trace("           %d directories per disk\n",
            skcache_bucket_num/radix_disk);

#if 0
        for(i=0;i<radix_disk;i++) {
            Trace("   disk# %3d: %s%d\n", i, cache_dir_prefix, i);
        }
#endif
    }
}

/*
 * initialized cache parameters and areas
 */
long long
skcache_init(time_t xmin_date, int xbk_num, int xsk_level, int xsk_len,
        int xffl_num, int xffl_size)
{
    int i,j;
    int chk;
    long long got;
    long long sum, tmp_sum;
    long long all;
    bucket_t *bk;
    time_t now;

    long long tmp;

#if 0
    Trace("skcache_init: min_date %d\n\
    bucket: number %d, level %d, length %d\n\
    frag: length %d, size %d\n",
        xmin_date, xbk_num, xsk_level, xsk_len, xffl_num, xffl_size);
#endif

    if(xmin_date>0) {
        min_date = xmin_date;
    }
    else {
#if 0
        min_date = time(NULL);
        Trace("set min_date as %d (current date/time)\n", min_date);
#endif
        Trace("set min_date as negative it is mean, use stored objects before boot\n");
    }

    now = time(NULL);

    sum = 0L;

    NIL = (skcache_entry_t*)malloc(sizeof(bucket_t));
    NIL->h1 = INT_MAX;
    NIL->h2 = INT_MAX;
    NIL->h3 = INT_MAX;
    MLOCK((void*)NIL, sizeof(bucket_t));

    sum += sizeof(bucket_t);
    mlist_add_elem("NIL", (void*)NIL, sizeof(bucket_t));

    skcache_bucket_num = xbk_num;

#if 0
    Trace("require_radix_h2 %d, require_radix_h3 %d\n",
            require_radix_h2, require_radix_h3);
#endif
    
    radix_h1 = skcache_bucket_num;
    radix_h2 = require_radix_h2;
    radix_h3 = require_radix_h3;
    radix_h4 = 7537;


    if(radix_h2>RADIX_H2_MAX) {
        Error("radix_h2 (%d) is too large (max %d)\n",
            radix_h2, RADIX_H2_MAX);
    }
    Trace("\n\
Hashes:\n\
  Radixes               %d, %d, %d, %d\n",
        radix_h1, radix_h2, radix_h3, radix_h4);

    skcache_bucket_base = (bucket_t*)malloc(sizeof(bucket_t)*xbk_num);
    if(skcache_bucket_base==NULL) {
        Error("not enough memory for skcache_bucket_base");
        skcache_bucket_num = 0;
        return -1;
    }
    MLOCK((void*)skcache_bucket_base, sizeof(bucket_t)*xbk_num);
    sum += sizeof(bucket_t)*xbk_num;
    mlist_add_elem("bucket bases",
        (void*)skcache_bucket_base, sizeof(bucket_t)*xbk_num);

    tmp_sum = 0LL;

    for(i=0;i<skcache_bucket_num;i++) {
#if 0
Trace("bk# %d %p\n", i, skcache_bucket_base[i]);
#endif

        chk = bk_new(&skcache_bucket_base[i], xsk_level, xsk_len, NIL);
        if(chk==0) {
            Error("initialize error in bucket# %d", i);
            return -2;
        }
        tmp_sum += chk;

        bk = &skcache_bucket_base[i];
        bk->bid = i;
        bk->own = '.';

        got = alloc_entry_freelist(&bk->entry_freelist, xsk_len+1);
        if(got<=0) {
            Error("initialize error in object entry elem_freelist# %d", i);
            return -2;
        }
        else {
            tmp_sum += got;
        }

        bk->entry_all  = xsk_len+1;
        bk->entry_use  = 0;
        bk->entry_free = xsk_len+1;
#if 0
        bk->entry_hhwater   = (int) (98.0 * bk->entry_all/100);
        bk->entry_hwater    = (int) (92.0 * bk->entry_all/100);
        bk->entry_lwater    = (int) (69.0 * bk->entry_all/100);
#endif
#if 1
        bk->entry_hhwater   = 98*bk->entry_all/100 - rand()%(bk->entry_all/20); 
        bk->entry_hwater    = 82*bk->entry_all/100 - rand()%(bk->entry_all/20); 
        bk->entry_lwater    = 69*bk->entry_all/100 - rand()%(bk->entry_all/20); 
#endif

#if 0
    Trace("bk# %4d: %5d %5d->%5d %5d->%5d\n",
        bk->bid, bk->entry_all,
        80*bk->entry_all/100, bk->entry_hwater,
        60*bk->entry_all/100, bk->entry_lwater
        );
#endif

#if 0
#ifndef NDEBUG
        bk->entry_first = (skcache_entry_t*) ((char*)bk->entry_freelist
                    + sizeof(skcache_entry_t)* 0);
        bk->entry_last = (skcache_entry_t*) ((char*)bk->entry_freelist
                    + sizeof(skcache_entry_t)* (xsk_len+1));
#endif

#ifndef NDEBUG
#if 1
    Trace(" %3d head of freelist %p first %p last %p\n",
        i, bk->entry_freelist, bk->entry_first, bk->entry_last);
    Trace("     2nd of freelist %p next of first %p\n",
        bk->entry_freelist->next, 
        (char*)bk->entry_freelist + sizeof(skcache_entry_t)* 1
    );
#endif
#endif
#endif

        bk->frag_use = 0;

        bk->gc_last = -1;
        bk->gc_push = 0; 
        bk->gc_rest = 0;
        bk->gc_checkpoint = now - (rand()% (skcache_mem_keep_cur/2) );
        bk->gc_ntarget = 0;

#ifdef USE_3LEVEL_DIR
        bk->dirpos = rand()%(radix_h2*radix_h3);
#else
        bk->dirpos = rand()%radix_h2;
#endif
        bk->dirty2 = -1;
        bk->dirty3 = -1;

    }

#if 1
    Trace("Memory:\n\
 entries:\n\
  the number of entries %d (/bucket)\n\
  watermark             hhigh %d (%.2f%%), high %d (%.2f%%), low %d (%.2f%%)\n",
            bk->entry_all,
            bk->entry_hhwater,  ((double)bk->entry_hhwater*100)/bk->entry_all,
            bk->entry_hwater,   ((double)bk->entry_hwater*100)/bk->entry_all,
            bk->entry_lwater,   ((double)bk->entry_lwater*100)/bk->entry_all
        );
#endif

    sum += tmp_sum;
    mlist_add_elem("bucket bodies", (void*)NULL, tmp_sum);

    got = alloc_frag_freelist(&z_frag_freelist, 
            (xffl_num+1)*skcache_bucket_num, xffl_size);
    if(got<0) {
        Error("Sorry, no memory for frag free list");
        return -2;
    }

    frag_free = (xffl_num+1)*skcache_bucket_num;
    frag_use  = 0;

#if 1
    frag_hwater = 16*frag_free/20;
    frag_lwater = 12*frag_free/20;
#endif
    skcache_frag_hwater_ratio = 80.0;
    skcache_frag_lwater_ratio = 60.0;

    sum += got;
    mlist_add_elem("frags", (void*)z_frag_freelist, got);

#if 1
    Trace(" frags:\n\
  the number of frags   %d\n\
  watermark             high %.2f%%, low %.2f%%\n",
        frag_free, skcache_frag_hwater_ratio, skcache_frag_lwater_ratio);
#endif

#ifdef USE_3LEVEL_DIR
    dir_num = radix_h1*radix_h2*radix_h3;
#else
    dir_num = radix_h1*radix_h2;
#endif

    if(dir_num<0) {
        Trace("dir_num %d is stange\n", dir_num);
        exit(4);
    }

    dir_size_max    = require_dir_size;
    dir_size_hwater = require_dir_size*9/10;
    dir_size_lwater = require_dir_size*6/10;

#if 1
    Trace("\nDisk:\n");
    Trace("  The number of directories      %d\n",
        dir_num);
    Trace("  Max size of each directory     %lld %.2fK %.2fM \n",
        (long long) dir_size_max, fdK(dir_size_max), fdM(dir_size_max));
    Trace("    High water                   %.2f%% %lld %.2fK %.2fM \n",
        ((double) dir_size_hwater*100.0)/dir_size_max,
        (long long) dir_size_hwater, fdK(dir_size_hwater), fdM(dir_size_hwater));
    Trace("    Low water                    %.2f%% %lld %.2fK %.2fM \n",
        ((double) dir_size_lwater*100.0)/dir_size_max,
        (long long) dir_size_lwater, fdK(dir_size_lwater), fdM(dir_size_lwater));
    tmp = ((long long) dir_size_max * dir_num);
    Trace("  Actual cache size on disk      %lld %.2fM %.2fG\n",
            tmp, fdM(tmp), fdG(tmp));
#endif

    check_disk_spliting();

    dir_lock = (int*)malloc(sizeof(int)*dir_num);
    MLOCK((void*)dir_lock, sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_lock",   (void*)dir_lock, sizeof(int)*dir_num);

#if 1
    Trace("\n\
Signatures for directories:
  SIZE_OF_UINT        %d\n\
  DSIG_BIT_WIDTH      %d\n\
  DSIG_VALUE_RANGE    %d\n\
  DSIG_VALUE_RES      %d\n\
  DSIG_ELEM_WIDTH     %d\n\
  DSIG_BYTE_WIDTH     %d\n",
        SIZE_OF_UINT,
        DSIG_BIT_WIDTH, DSIG_VALUE_RANGE, DSIG_VALUE_RES,
        DSIG_ELEM_WIDTH, DSIG_BYTE_WIDTH);
    Trace("\n");
#endif

#if 0
    dir_sig = (int*)malloc(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_sig",   (void*)dir_sig, sizeof(int)*dir_num);
#endif
#if 0
    dir_sig = (int64_t*)malloc(sizeof(int64_t)*dir_num);
    MLOCK((void*)dir_sig, sizeof(int64_t)*dir_num);
    sum += sizeof(int64_t)*dir_num;
    mlist_add_elem("dir_sig",   (void*)dir_sig, sizeof(int64_t)*dir_num);
#endif
    dir_sig = (u_int32_t*)malloc(sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);
    MLOCK((void*)dir_sig, sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);
    sum += sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH;
    mlist_add_elem("dir_sig",   (void*)dir_sig, sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);

    dir_sz  = (off_t*)malloc(sizeof(off_t)*dir_num);
    MLOCK((void*)dir_sz, sizeof(off_t)*dir_num);
    sum += sizeof(off_t)*dir_num;
    mlist_add_elem("dir_sz",    (void*)dir_sz, sizeof(off_t)*dir_num);

    dir_count = (int*)malloc(sizeof(int)*dir_num);
    MLOCK((void*)dir_count, sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_count", (void*)dir_count, sizeof(int)*dir_num);

    dir_time = (int*)malloc(sizeof(int)*dir_num);
    MLOCK((void*)dir_time, sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_time",  (void*)dir_time, sizeof(int)*dir_num);

    for(i=0;i<dir_num;i++) {
#if 0
        dir_sig[i]      = (int64_t)-1L;
#endif
        DSIG_ALLSET(i);
        dir_sz[i]       = (off_t)-1L;
        dir_count[i]    = -1;
        dir_time[i]     = now - rand()%skcache_signature_refresh;
        dir_lock[i]     = 0;
    }


    skcache_keeptime_token_init = require_keeptime_token_val;
    if(skcache_keeptime_token_init<2) {
        skcache_keeptime_token_init = 2;
    }
    skcache_keeptime_token = skcache_keeptime_token_init;

    Trace("  keeptime updating token    %d\n", skcache_keeptime_token_init);
    Trace("  keeptime updating interval %d\n", skcache_keeptime_token_init*2);


    swapout_token = swapout_token_init = require_swap_token_val;
    swapout_token_interval = require_swap_token_interval;


    Trace("swapout token %d in %d msec -> rate %.4f swapout/sec\n",
        require_swap_token_val, require_swap_token_interval,
        (double)require_swap_token_val*
            (1000.0/(double)require_swap_token_interval)
        );

    disk_sweep_interval = require_disk_sweep_interval;

#if 0
    Trace("skcache_init: allocate %lld byte(s) %dK %.2fM %.2fG\n",
        (long long) sum, (long)dK(sum), fdM(sum), fdG(sum));
#endif

    sim_stat_alloc_new(&swapload_time,  "swapload time [msec]",
        K(1),   swapload_time_count);
    sim_stat_alloc_new(&swapin_time,    "swapin time [msec]",
        K(1),   swapin_time_count);
    sim_stat_alloc_new(&swapout_time,   "swapout time [msec]",
        K(1),   swapout_time_count);

    sim_stat_U_alloc_new(&compaction_time,  "compaction time [msec]",
        100, K(2),   compaction_time_count);
    sim_stat_U_alloc_new(&checkdir_time,    "checkdir time [msec]",
        10, K(2),   checkdir_time_count);
    sim_stat_U_alloc_new(&checkdir_time_short,    "checkdir time (short) [msec]",
        10, K(2),   checkdir_time_short_stat_count);
    sim_stat_U_alloc_new(&checkmem_time,    "checkmem time [msec]",
        100, K(2),   checkmem_time_count);

    sim_stat_alloc_new(&swapout_fail_reason,    "swapout fail reason",
        K(1),   swapout_fail_reason_count);
    sim_stat_alloc_new(&swapload_fail_reason,   "swapload fail reason",
        K(1),   swapload_fail_reason_count);

    sim_hist_alloc_new(&fratio_hist,            "frag ratio history",
        skcache_keeptime_token_init*2,  fratio_hist_area);

    return sum;
}

